#!/usr/bin/env python3
import logging
import os
from enrichm.databases import Databases
from enrichm.data import Data
from enrichm.toolbox import run_command
################################################################################
class Plot:

    def __init__(self):
        self.databases = Databases()
        path_to_scripts = os.path.split(os.path.realpath(__file__))[0]
        self.draw_pca_script_path = os.path.join(path_to_scripts, "PLOT_ko_pca.r")
        self.draw_heatmap_script_path = os.path.join(path_to_scripts, "PLOT_ko_heatmap.r")
        self.draw_barplots_script_path = os.path.join(path_to_scripts, "PLOT_ko_breakdown.r")
        self.ko00000 = os.path.join(Data.DATABASE_DIR, self.databases.DB_VERSION, 'ko00000.tsv')
        self.output_pca_plot = 'presence_absence_pca_plot.svg'
        self.output_heatmap_plot = 'presence_absence_pca_plot.svg'

    def draw_pca_plot(self, annotation_matrix, metadata, output_directory):
        logging.info('	- Generating PCA plot')
        output_path = os.path.join(output_directory, self.output_pca_plot)

        cmd = f"Rscript {self.draw_pca_script_path} \
                    -i {annotation_matrix} \
                    -m {metadata} \
                    -o {output_path} > /dev/null 2>&1"
        run_command(cmd)

    def draw_barplots(self, annotation_matrix, pvalue, output_directory):
        logging.info('	- Generating KO breakdown plots')
        cmd = f"Rscript {self.draw_barplots_script_path} \
                    -i {annotation_matrix} \
                    -o {output_directory} \
                    -k {self.ko00000} \
                    -p {pvalue} > /dev/null 2>&1"
        run_command(cmd)



#!/usr/bin/env python
# Imports
import unittest
import os.path
import sys
import subprocess
import tempfile

path_to_script = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'bin', 'enrichm')
path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')]+sys.path

from enrichm.annotate import Annotate

###############################################################################

class Tests(unittest.TestCase):

    def test_hello_world_nucleic_dir(self):
        tmp = tempfile.mkdtemp()
        bin = os.path.join(path_to_data, 'test_nucleic_bin')
        cmd = '%s annotate \
                        --threads 10 \
                        --ko \
                        --pfam \
                        --tigrfam \
                        --genome_directory %s \
                        --output %s \
                        --force' % (path_to_script, bin, tmp)
        subprocess.call(cmd, shell=True)

    def test_hello_world_protein_dir(self):
        tmp = tempfile.mkdtemp()
        bin = os.path.join(path_to_data, 'test_protein_bin')

        cmd = '%s annotate \
                        --threads 10 \
                        --ko \
                        --pfam \
                        --tigrfam \
                        --protein_directory %s \
                        --output %s \
                        --force' % (path_to_script, bin, tmp)
        subprocess.call(cmd, shell=True)

    def test_hello_world_nucleic_file(self):
        tmp = tempfile.mkdtemp()
        genome_file = os.path.join(path_to_data, 'test_nucleic_bin', "GCF_001889405.1_ASM188940v1_subset.fna")
        cmd = '%s annotate \
                        --threads 10 \
                        --ko \
                        --pfam \
                        --tigrfam \
                        --genome_files %s \
                        --output %s \
                        --force' % (path_to_script, genome_file, tmp)
        subprocess.check_call(cmd, shell=True)

    def test_hello_world_protein_file(self):
        tmp = tempfile.mkdtemp()
        protein_file = os.path.join(path_to_data, 'test_protein_bin', "GCF_001889405.1_ASM188940v1_subset.faa")

        cmd = '%s annotate \
                        --threads 10 \
                        --ko \
                        --pfam \
                        --tigrfam \
                        --protein_files %s \
                        --output %s \
                        --force' % (path_to_script, protein_file, tmp)
        subprocess.check_call(cmd, shell=True)

    def test_hello_world_hypothetical_cluster(self):
        tmp = tempfile.mkdtemp()
        protein_file = os.path.join(path_to_data, 'test_protein_bin', "GCF_001889405.1_ASM188940v1_subset.faa")

        cmd = '%s annotate \
                        --verbosity 5 \
                        --threads 4 \
                        --clusters \
                        --protein_files %s \
                        --output %s \
                        --force' % (path_to_script, protein_file, tmp)
        subprocess.check_call(cmd, shell=True)

    def test_very_simple_orthology(self):
        tmp = tempfile.mkdtemp()
        protein_file = os.path.join(path_to_data, 'cluster_data', "very_simple_test.faa")

        cmd = '%s annotate \
                        --threads 4 \
                        --orthologs \
                        --protein_files %s \
                        --output %s \
                        --force' % (path_to_script, protein_file, tmp)
        subprocess.check_call(cmd, shell=True)

    def test_very_simple_homology(self):
        tmp = tempfile.mkdtemp()
        protein_file = os.path.join(path_to_data, 'cluster_data', "very_simple_test.faa")

        cmd = '%s annotate \
                        --threads 4 \
                        --cluster \
                        --protein_files %s \
                        --output %s \
                        --force' % (path_to_script, protein_file, tmp)
        subprocess.check_call(cmd, shell=True)

    def test(self):
        tmp = tempfile.mkdtemp()
        self.simple_annotate_instance \
            = Annotate(tmp,
                       True, True, True, True, True, True, True, True, True,# Annotate with all databases
                       1e-05, 0, 0.3, 0.7, 0.7, 0.7, # Runtime options
                       False, False, False, False, False, False, True, # Cutoffs
                       5, # Inflation
                       4, 2500, # chunks
                       False, 1, 1, '.fna', False)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# Imports
import unittest
import os
import sys
import tempfile
import filecmp
import shutil

path_to_script = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'bin', 'enrichm')
path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')]+sys.path

from enrichm.classify_checks import ClassifyChecks
from enrichm.parser import RulesJson, RulesJsonVersion0, Parser

###############################################################################

class Tests(unittest.TestCase):

    def test_basic(self):
        parser = Parser
        rules = RulesJson()
        basic_rules = os.path.join(path_to_data, "rules/basic_rules_1.json")
        basic_gff = os.path.join(path_to_data, "gffs/basic_gff_1.gff")

        basic_rules_object = rules.load(basic_rules)
        classify_checks = ClassifyChecks(basic_rules_object)
        features, _ =  parser.parse_gff(basic_gff)

        self.assertFalse(classify_checks.check("sample_pathway", features['genome_1']))
        self.assertFalse(classify_checks.check("sample_pathway", features['genome_2']))
        self.assertTrue(classify_checks.check("sample_pathway", features['genome_3']))

    def test_split(self):
        parser = Parser
        rules = RulesJson()
        basic_rules = os.path.join(path_to_data, "rules/basic_rules_2.json")
        basic_gff = os.path.join(path_to_data, "gffs/basic_gff_2.gff")
        basic_rules_object = rules.load(basic_rules)
        classify_checks = ClassifyChecks(basic_rules_object)
        features, _ =  parser.parse_gff(basic_gff)

        self.assertTrue(classify_checks.check("sample_pathway", features['genome_1']))
        self.assertFalse(classify_checks.check("sample_pathway", features['genome_2']))

    def test_missing_genes(self):
        parser = Parser
        rules = RulesJson()
        basic_rules = os.path.join(path_to_data, "rules/basic_rules_3.json")
        basic_gff = os.path.join(path_to_data, "gffs/basic_gff_3.gff")
        basic_rules_object = rules.load(basic_rules)
        classify_checks = ClassifyChecks(basic_rules_object)
        features, _ =  parser.parse_gff(basic_gff)

        self.assertTrue(classify_checks.check("sample_pathway", features['genome_1']))
        self.assertFalse(classify_checks.check("sample_pathway", features['genome_2']))

    @unittest.skip("Known failure")
    def test_split_contig(self):
        parser = Parser
        rules = RulesJson()
        basic_rules = os.path.join(path_to_data, "rules/basic_rules_4.json")
        basic_gff = os.path.join(path_to_data, "gffs/basic_gff_4.gff")
        basic_rules_object = rules.load(basic_rules)
        classify_checks = ClassifyChecks(basic_rules_object)
        features, _ =  parser.parse_gff(basic_gff)

        self.assertTrue(classify_checks.check("sample_pathway", features['genome_1']))

    # TODO: Check overlapping contig names

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=line-too-long
"""
Various functions for annotating genomes.
"""
# System imports

import pickle
import shutil
import os
import tempfile
import logging
import subprocess
import multiprocessing as mp
from os import path, close, mkdir, listdir
from enrichm.genome import Genome, AnnotationParser
from enrichm.databases import Databases
from enrichm.sequence_io import SequenceIO
from enrichm.writer import Writer, MatrixGenerator
from enrichm.toolbox import list_splitter, run_command

def parse_genomes(params):
    '''
    Parses an input genome file into a Genome object. This is outside
    of a class to enable parallelisation
    '''

    genome = Genome(*params)
    return genome

class Annotate:
    '''
    Annotates proteins, and MAGs
    '''
    GENOME_BIN = 'genome_bin'
    GENOME_PROTEINS = 'genome_proteins'
    GENOME_GENES = 'genome_genes'
    GENOME_KO = 'annotations_ko'
    GENOME_KO_HMM = 'annotations_ko_hmm'
    GENOME_EC = 'annotations_ec'
    GENOME_PFAM = 'annotations_pfam'
    GENOME_TIGRFAM = 'annotations_tigrfam'
    GENOME_HYPOTHETICAL = 'annotations_hypothetical'
    GENOME_CAZY = 'annotations_cazy'
    GENOME_GFF = 'annotations_gff'
    GENOME_OBJ = 'annotations_genomes'
    OUTPUT_KO = 'ko_frequency_table.tsv'
    OUTPUT_KO_HMM = 'ko_hmm_frequency_table.tsv'
    OUTPUT_EC = 'ec_frequency_table.tsv'
    OUTPUT_PFAM = 'pfam_frequency_table.tsv'
    OUTPUT_TIGRFAM = 'tigrfam_frequency_table.tsv'
    OUTPUT_CAZY = 'cazy_frequency_table.tsv'
    OUTPUT_CLUSTER = 'cluster_frequency_table.tsv'
    OUTPUT_ORTHOLOG = 'ortholog_frequency_table.tsv'
    OUTPUT_HYPOTHETICAL_ANNOTATIONS = 'hypothetical_annotations.tsv'
    OUTPUT_DIAMOND = "DIAMOND_search"
    GFF_SUFFIX = '.gff'
    PROTEINS_SUFFIX = '.faa'
    ANNOTATION_SUFFIX = '.tsv'
    PICKLE_SUFFIX = '.pickle'

    def __init__(self, output_directory, annotate_ko, annotate_ko_hmm, annotate_pfam,
                 annotate_tigrfam, annoatate_cluster, annotate_ortholog, annotate_cazy, annotate_ec,
                 annotate_orthogroup, evalue, bit, percent_id_cutoff, aln_query, aln_reference, 
                 fraction_aligned, cut_ga_pfam, cut_nc_pfam, cut_tc_pfam, cut_ga_tigrfam, cut_nc_tigrfam,
                 cut_tc_tigrfam, cut_hmm, inflation, chunk_number, chunk_max,
                 count_domains, threads, parallel, suffix, light):


        # Define inputs and outputs
        self.output_directory = output_directory

        # Define type of annotation to be carried out
        self.annotate_ko = annotate_ko
        self.annotate_ko_hmm = annotate_ko_hmm
        self.annotate_pfam = annotate_pfam
        self.annotate_tigrfam = annotate_tigrfam
        self.annotate_cluster = annoatate_cluster
        self.annotate_ortholog = annotate_ortholog
        self.annotate_orthogroup = annotate_orthogroup
        self.annotate_cazy = annotate_cazy
        self.annotate_ec = annotate_ec

        # Cutoffs
        self.evalue = evalue
        self.bit = bit
        self.percent_id_cutoff = percent_id_cutoff
        self.aln_query = aln_query
        self.aln_reference = aln_reference
        self.fraction_aligned = fraction_aligned
        self.cut_ga_pfam = cut_ga_pfam
        self.cut_nc_pfam = cut_nc_pfam
        self.cut_tc_pfam = cut_tc_pfam
        self.cut_ga_tigrfam = cut_ga_tigrfam
        self.cut_nc_tigrfam = cut_nc_tigrfam
        self.cut_tc_tigrfam = cut_tc_tigrfam
        self.cut_hmm = cut_hmm
        self.inflation = inflation
        self.chunk_number = chunk_number
        self.chunk_max = chunk_max
        self.count_domains = count_domains

        # Parameters
        self.threads = threads
        self.parallel = parallel
        self.suffix = suffix
        self.light = light

        # Set up multiprocesses pool
        self.pool = mp.Pool(processes=int(self.parallel))

        # Load databases
        self.databases = Databases()

    def prep_genome(self, genome_file_list, genome_directory):
        '''
        Do any preparation specific to the genome annotation pipeline.

        Inputs
        ------
        genome_file_list - List. list of strings, each a path to a file
        containing a genome

        Outputs
        -------
        returns the directory with all genome ids sym-linked into it.
        '''
        # link all the genomes into one file
        logging.info('Preparing genomes for annotation')

        if genome_file_list:
            mkdir(genome_directory)
            genome_paths = list()

            for genome_path in genome_file_list:

                if genome_path.endswith(self.suffix):
                    genome_paths.append(f"{genome_path}")

            cmd = f"xargs --arg-file=/dev/stdin cp --target-directory={genome_directory}"

            logging.debug(cmd)
            process = subprocess.Popen(["bash", "-c", cmd],
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       universal_newlines=True)
            process.communicate(input=str('\n'.join(genome_paths)))

        return genome_directory

    def call_proteins(self, genome_directory):
        '''
        Use prodigal to call proteins within the genomes

        Parameters
        ----------
        genome_directory  - string. Directory containing .fna files for each
                            input genome

        Outputs
        -------
        returns the directory containing an .faa file for each input genomes
        '''
        protein_directory_path = path.join(self.output_directory, self.GENOME_PROTEINS)
        gene_directory_path = path.join(self.output_directory, self.GENOME_GENES)
        mkdir(protein_directory_path)
        mkdir(gene_directory_path)
        genome_list = list()
        genome_paths = list()

        for genome in listdir(genome_directory):

            if genome.endswith(self.suffix):
                genome_paths.append(path.splitext(genome)[0])

        logging.info("    - Calling proteins for %i genomes", len(genome_paths))
        cmd = "ls %s/*%s | \
                    sed 's/%s//g' | \
                    grep -o '[^/]*$' | \
                    parallel -j %s \
                        prodigal \
                            -q \
                            -p meta \
                            -o /dev/null \
                            -d %s/{}%s \
                            -a %s/{}%s \
                            -i %s/{}%s \
                            > /dev/null 2>&1" \
                % (genome_directory, self.suffix, self.suffix, self.parallel, gene_directory_path,
                   self.suffix, protein_directory_path, self.PROTEINS_SUFFIX, genome_directory,
                   self.suffix)

        run_command(cmd)

        protein_directory_files = listdir(protein_directory_path)
        genome_directory_files = listdir(genome_directory)

        for genome_protein, genome_nucl in zip(protein_directory_files, genome_directory_files):
            genome_protein_base = genome_protein.replace(self.PROTEINS_SUFFIX, self.suffix)
            output_genome_protein_path = path.join(protein_directory_path, genome_protein)
            output_genome_nucl_path = path.join(genome_directory, genome_nucl)
            output_genome_gene_path = path.join(gene_directory_path, genome_protein_base)

            genome = (self.light, output_genome_protein_path, output_genome_nucl_path,
                      output_genome_gene_path)
            genome_list.append(genome)

        return genome_list

    def annotate_diamond(self, genomes_list, database, parser_type, ids_type, output_subdirectory):
        '''
        Annotate the proteins encoded by each genome with KO ids using either BLAST or using HMM
        searches (no implemented yet).

        Parameters
        ----------
        genome_faa_directory  - string. Directory containing .faa files for
                                each input genome

        Outputs
        -------
        returns a directory containing the search results for each of the input population genomes,
        and a frequency matrix contining with the KOs as rows, and the genomes as columns.
        '''

        output_directory_path = path.join(self.output_directory,
                                          output_subdirectory)
        genome_dict = {genome.name:genome for genome in genomes_list}
        mkdir(output_directory_path)
        specific_cutoffs = None

        with tempfile.NamedTemporaryFile() as temp:

            to_write = str()

            for genome in genomes_list:
                to_write += f"sed \"s/>/>{genome.name}~/g\" {genome.path}\n"

            temp.write(str.encode(to_write))
            temp.flush()
            output_annotation_path = path.join(output_directory_path, self.OUTPUT_DIAMOND) + \
                                        self.ANNOTATION_SUFFIX
            logging.info('    - BLASTing genomes')
            self.diamond_search(temp.name, output_annotation_path, database)

            for genome_name, batch in self.get_batches(output_annotation_path):

                if batch:
                    genome = genome_dict[genome_name]
                    genome.add(batch, self.evalue, self.bit, self.aln_query, self.aln_reference,
                               specific_cutoffs, parser_type, ids_type)

    def get_batches(self, input_file):
        '''
        Separate DIAMOND blast results into batches, where a batch is all the hits for a genome.

        Parameters
        ----------
        input_file - string. Directory to search for blast results.
        '''

        last = None
        input_file_io = open(input_file)

        for line in input_file_io:
            split_line = line.strip().split('\t')
            genome_id = split_line[0].split('~')[0]

            if last is None:
                last = genome_id
                batch = [split_line]

            else:

                if last == genome_id:
                    batch.append(split_line)
                else:
                    yield last, batch
                    batch = [split_line]
                    last = genome_id

        if last is None:
            yield None, None
        else:
            yield last, batch

    def diamond_search(self, tmp_name, output_path, database):
        '''
        Carry out a diamond blastp search.

        Parameters
        ----------
        input_genome_path - string. Path to file containing .faa file for an input genome
        output_path - string. Path to file to output results into
        databases - string. Path to HMM to use for searching
        '''

        cmd = f'bash {tmp_name} | diamond blastp \
                                    --quiet \
                                    --outfmt 6 \
                                    --max-target-seqs 1 \
                                    --query /dev/stdin \
                                    --out {output_path} \
                                    --db {database} \
                                    --threads {self.threads} '
        if self.evalue:
            cmd += f'--evalue {self.evalue} '

        if self.bit:
            cmd += f'--min-score {self.bit} '

        if self.percent_id_cutoff:
            cmd += f'--id {self.percent_id_cutoff*100} '

        if self.aln_query:
            cmd += f"--query-cover {self.aln_query*100} "

        if self.aln_reference:
            cmd += f"--subject-cover {self.aln_reference*100} "

        run_command(cmd)

    def hmmsearch_annotation(self, genomes_list, output_directory_path, database, ids_type, parser):
        '''
        Annotate the proteins encoded by each genome with pfam ids using HMM searches.

        Parameters
        ----------
        genomes_list - list. list of Genome objects

        '''
        mkdir(output_directory_path)
        genome_dict = {genome.name: genome for genome in genomes_list}

        hmmcutoff = (ids_type in (AnnotationParser.TIGRFAM, AnnotationParser.PFAM))

        if ids_type == AnnotationParser.KO_HMM:
            specific_cutoffs = self.databases.parse_ko_cutoffs()
        else:
            specific_cutoffs = None

        self.hmm_search(output_directory_path, database, hmmcutoff)

        if ids_type == AnnotationParser.PFAM:
            pfam2clan = self.databases.pfam2clan()
        else:
            pfam2clan = None

        for genome_annotation in listdir(output_directory_path):
            genome_id = path.splitext(genome_annotation)[0]
            genome = genome_dict[genome_id]
            output_annotation_path = path.join(output_directory_path, genome_annotation)
            genome.add(output_annotation_path, self.evalue, self.bit, self.aln_query,
                       self.aln_reference, specific_cutoffs, parser, ids_type,
                       pfam2clan=pfam2clan)

    def annotate_hypothetical(self, genomes_list):
        '''
        Sort proteins coded by each genome into homologous clusters.

        Inputs
        ------
        genomes_list - list. list of Genome objects

        '''
        output_directory_path = path.join(self.output_directory, self.GENOME_HYPOTHETICAL)
        mkdir(output_directory_path)

        renamed_genomes = list()
        for genome in genomes_list:
            renamed_genome = next(tempfile._get_candidate_names())
            cmd = f"sed 's/>/>{genome.name}~/g' {genome.path} > {renamed_genome}"
            run_command(cmd)
            renamed_genomes.append(renamed_genome)

        
        tmp_dir = tempfile.mkdtemp()

        db_path = path.join(output_directory_path, "db")
        clu_path = path.join(output_directory_path, "clu")
        align_path = path.join(output_directory_path, "alignDb")
        blast_output_path = path.join(output_directory_path, "alignDb.m8")
        formatted_blast_output_path = path.join(output_directory_path, "alignDb.formatted.m8")

        clu_tsv_path = path.join(output_directory_path, "hypothetical_clusters.tsv")

        logging.info('    - Generating MMSeqs2 database')
        cmd = f"mmseqs createdb {' '.join(renamed_genomes)} {db_path}"
        run_command(cmd)
        for renamed_genome in renamed_genomes:
            os.remove(renamed_genome)

        logging.info('    - Clustering genome proteins')
        cmd = f"mmseqs cluster \
                    {db_path} \
                    {clu_path} \
                    {tmp_dir} \
                    --threads {self.threads} \
                    --min-seq-id {self.percent_id_cutoff} \
                    -c {self.fraction_aligned} \
                    -v 0"
        run_command(cmd)

        logging.info('    - Extracting clusters')
        cmd = f'mmseqs createtsv \
                    {db_path} \
                    {db_path} \
                    {clu_path} \
                    {clu_tsv_path} \
                    --threads {self.threads} \
                    -v 0'
        run_command(cmd)

        if self.annotate_ortholog:

            logging.info('    - Computing Smith-Waterman alignments for clustering results')
            cmd = f"mmseqs alignall \
                        {db_path} \
                        {clu_path} \
                        {align_path} \
                        --alignment-mode 3 \
                        --threads {self.threads} \
                        -v 0"
            run_command(cmd)

            logging.info('    - Converting to BLAST-like output')
            cmd = f"mmseqs createtsv \
                        {db_path} \
                        {db_path} \
                        {align_path} \
                        {blast_output_path} \
                        --threads {self.threads} \
                        -v 0"
            # --format-output query,target,bits
            run_command(cmd)

            logging.info('    - Reformatting BLAST output')
            cmd = "OFS=\"\t\" awk 'FNR==NR{a[$1]=$2;next}{$3=a[$3]; \
                                            $1=\"\"; for(i=2;i<NF;i++){printf(\"%s\t\",$i)} \
                                            printf(\"\\n\")}' %s %s | cut -f1,2,5 > %s" \
                % ("%s", db_path + '.lookup', blast_output_path, formatted_blast_output_path)
            run_command(cmd)

            ortholog_dict = self.run_mcl(formatted_blast_output_path,
                                            output_directory_path)
            ortholog_ids = ortholog_dict.keys()
        else:
            ortholog_dict = dict()
            ortholog_ids = list()
        cluster_ids = self.parse_cluster_results(clu_tsv_path,
                                                 genomes_list,
                                                 ortholog_dict,
                                                 output_directory_path)
        return cluster_ids, ortholog_ids

    def run_mcl(self, blast_abc, output_directory_path):
        '''
        Parse the protein clusters produced from Mmseqs2 using mcl

        Parameters
        ----------
        blast_abc - string. an abc file for mcl to run on. More information on the format of abc
                    files can be found at https://micans.org/mcl/man/clmprotocols.html
        output_directory_path - string. Path to write the results of mcl parsing to.
        '''

        dict_path = path.join(output_directory_path, "alignDb.dict")
        mci_path = path.join(output_directory_path, "alignDb.mci")
        cluster_path = path.join(output_directory_path, "mcl_clusters.tsv")
        output_path = path.join(output_directory_path, "mcl_clusters.convert.tsv")
        
        logging.info('    - Preparing network')
        ortholog_dict = dict()
        cmd = f"mcxload \
                    -abc {blast_abc} \
                    -write-tab {dict_path} \
                    -o {mci_path} \
                    --stream-mirror \
                    --stream-neg-log10"
        run_command(cmd)
        logging.info('    - Finding orthologs')
        ortholog_dict = dict()
        cmd = f'mcl \
                    {mci_path} \
                    -te {self.threads} \
                    -I {self.inflation} \
                    -o {cluster_path}'
        run_command(cmd)

        logging.info('    - Reformatting output')
        ortholog_dict = dict()
        cmd = f'mcxdump \
                    -icl {cluster_path} \
                    -o {output_path} \
                    -tabr {dict_path}'
        run_command(cmd)

        ortholog = 1
        for line in open(output_path):
            ortholog_idx = "ortholog_%i" % ortholog
            ortholog_dict[ortholog_idx] = set()

            for protein in line.strip().split('\t'):
                ortholog_dict[ortholog_idx].add(protein)

            ortholog += 1
        return ortholog_dict

    def parse_cluster_results(self, cluster_output_path, genomes_list, ortholog_dict,
                              output_directory_path):
        '''
        Parse cluster output in tab format.

        Inputs
        ------
        from_cluster_results    - String. Path to mmseqs2 clustering output file

        Yields
        -------
        A cluster name, and a list of sequences in that cluster.

        '''
        logging.info('    - Parsing input cluster file: %s', cluster_output_path)

        cluster_ids = set()
        previous_cluster_name = None
        counter = 0
        genome_dictionary = {genome.name:genome for genome in genomes_list}

        output_hypothetical_annotations = path.join(output_directory_path,
                                                    self.OUTPUT_HYPOTHETICAL_ANNOTATIONS)
        with open(output_hypothetical_annotations, 'w') as out_io:

            for line in open(cluster_output_path):

                cluster_id, member = line.strip().split('\t')
                genome_id, sequence_id = member.split('~')

                if cluster_id == previous_cluster_name:
                    genome_dictionary[genome_id].add_cluster(sequence_id, "cluster_%i" % counter)
                else:
                    counter += 1
                    previous_cluster_name = cluster_id
                    cluster_ids.add("cluster_%i" % counter)
                    genome_dictionary[genome_id].add_cluster(sequence_id, "cluster_%i" % counter)

                out_io.write('\t'.join([genome_id, sequence_id, "cluster_%i" % counter]) + '\n')

        for ortholog, group in ortholog_dict.items():

            for member in group:
                genome, protein = member.split('~')
                genome_dictionary[genome].add_ortholog(protein, ortholog)

        return cluster_ids

    def _default_hmmsearch_options(self):
        cmd = str()

        if self.bit:
            cmd += '-T %s ' % (str(self.bit))
        else:
            cmd += '-E %s ' % (str(self.evalue))

        return cmd

    def hmm_search(self, output_path, database, hmmcutoff):
        '''
        Carry out a hmmsearch.

        Parameters
        ----------
        input_genome_path     - string. Path to file containing .faa file for
                                an input genome
        output_path           - string. Path to file to output results into
        databases             - string. Path to HMM to use for searching
        '''

        input_genome_path = path.join(self.output_directory, self.GENOME_PROTEINS)
        cmd = "ls %s | sed 's/%s//g' | parallel -j %s\
                                                hmmsearch \
                                                    --cpu %s \
                                                    -o /dev/null \
                                                    --noali \
                                                    --domtblout %s/{}%s " \
                          % (input_genome_path, self.PROTEINS_SUFFIX, self.parallel,
                             self.threads, output_path, self.ANNOTATION_SUFFIX)
        if hmmcutoff:
            if (self.cut_ga_pfam or self.cut_nc_pfam or self.cut_tc_pfam) and 'pfam' in database:
                if self.cut_ga_pfam:
                    cmd += " --cut_ga "
                if self.cut_nc_pfam:
                    cmd += " --cut_nc "
                if self.cut_tc_pfam:
                    cmd += " --cut_tc "
            elif (self.cut_ga_tigrfam or self.cut_nc_tigrfam or self.cut_tc_tigrfam) and 'tigrfam' in database:
                if self.cut_ga_tigrfam:
                    cmd += " --cut_ga "
                if self.cut_nc_tigrfam:
                    cmd += " --cut_nc "
                if self.cut_tc_tigrfam:
                    cmd += " --cut_tc "
            else:
                cmd += self._default_hmmsearch_options()
        else:
            cmd += self._default_hmmsearch_options()

        cmd += "%s %s/{}.faa 2> /dev/null" % (database, input_genome_path)

        run_command(cmd)

    def generate_gff_files(self, genomes_list):
        '''
        Write GFF files for each of the genome objects in genomes_list

        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        output_directory_path = path.join(self.output_directory,
                                          self.GENOME_GFF)
        mkdir(output_directory_path)
        for genome in genomes_list:
            logging.info('    - Generating .gff file for %s', genome.name)
            gff_output = path.join(output_directory_path, genome.name + self.GFF_SUFFIX)
            Writer.write_gff(genome, gff_output)

    def rename_fasta(self, genomes_list):
        '''
        Rename the called proteins with annotation ids.

        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        seqio = SequenceIO()

        for genome in genomes_list:
            file_object, fname = tempfile.mkstemp(suffix='.faa', text=True)

            if genome.gene:
                fd_gene, fname_gene = tempfile.mkstemp(suffix='.fna', text=True)

                with open(fname_gene, 'w') as out_gene_io:

                    for description, sequence in seqio.each(open(genome.gene)):
                        name = description.partition(' ')[0]
                        annotations = ' '.join(genome.sequences[name].all_annotations())
                        out_gene_io.write(">%s %s\n" % (name, annotations))
                        out_gene_io.write(sequence + '\n')

                close(fd_gene)
                logging.debug('Moving %s to %s', fname_gene, genome.gene)
                shutil.move(fname_gene, genome.gene)

            with open(fname, 'w') as out_io:

                for description, sequence in seqio.each(open(genome.path)):
                    name = description.partition(' ')[0]
                    annotations = ' '.join(genome.sequences[name].all_annotations())
                    out_io.write(">%s %s\n" % (name, annotations))
                    out_io.write(str(sequence) + '\n')

            close(file_object)
            logging.debug('Moving %s to %s', fname, genome.path)
            shutil.move(fname, genome.path)

    def pickle_objects(self, genomes_list):
        '''
        Store annotated genome objects as pickles.

        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        output_directory_path = path.join(self.output_directory,
                                          self.GENOME_OBJ)
        mkdir(output_directory_path)

        for genome in genomes_list:
            genome_pickle_path = path.join(output_directory_path, genome.name + self.PICKLE_SUFFIX)

            with open(genome_pickle_path, 'wb') as output:
                pickle.dump(genome, output)

    def parse_genome_inputs(self, genome_directory, protein_directory, genome_files, protein_files):
        '''
        Inputs
        ------

        Outputs
        -------

        '''

        prep_genomes_list = list()
        genomes_list = list()

        if protein_directory:
            logging.info("Using provided proteins")
            protein_genome_list = list()

            for protein_file in listdir(protein_directory):
                protein_genome_list.append(path.join(protein_directory, protein_file))

            directory = self.prep_genome(protein_genome_list,
                                         path.join(self.output_directory,
                                                   self.GENOME_PROTEINS))

            for genome_proteins_file in listdir(directory):

                if genome_proteins_file.endswith(self.suffix):
                    genome = (self.light, path.join(directory, genome_proteins_file), None, None)
                    prep_genomes_list.append(genome)

        elif protein_files:
            logging.info("Using provided proteins")
            genome_proteins_path = path.join(self.output_directory, self.GENOME_PROTEINS)
            directory = self.prep_genome(protein_files, genome_proteins_path)

            for protein_file in listdir(directory):
                protein_file_path = path.join(directory, path.basename(protein_file))
                prep_genomes_list.append((self.light, protein_file_path, None, None))

        elif genome_directory:
            logging.info("Calling proteins for annotation")
            prep_genomes_list = self.call_proteins(genome_directory)
            directory = genome_directory

        elif genome_files:
            logging.info("Calling proteins for annotation")
            directory = self.prep_genome(genome_files,
                                         path.join(self.output_directory, self.GENOME_BIN))
            prep_genomes_list = self.call_proteins(directory)

        for chunk in list_splitter(prep_genomes_list, self.chunk_number, self.chunk_max):
            genomes_list += self.pool.map(parse_genomes, chunk)

        return genomes_list

    def annotate_pipeline(self, genome_directory, protein_directory, genome_files, protein_files):
        '''
        Run Annotate pipeline for enrichM

        Parameters
        ----------
        genome_directory    - String. Path to directory containing genomes
        protein_directory   - String. Path to directory containing proteins (.faa files) for genomes
        genome_files        - List. List of strings, each to a .fna genome file.
        protein_files       - List. List of strings, each to a .faa proteins file.
        '''

        logging.info("Running pipeline: annotate")
        logging.info("Setting up for genome annotation")
        genomes_list = self.parse_genome_inputs(genome_directory, protein_directory,
                                                genome_files, protein_files)

        if genomes_list:
            logging.info("Starting annotation:")

            if (self.annotate_cluster or self.annotate_ortholog):
                logging.info('    - Annotating genomes with hypothetical clusters')
                cluster_ids, ortholog_ids = self.annotate_hypothetical(genomes_list)

                logging.info('    - Generating hypotheticals frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.HYPOTHETICAL, cluster_ids)
                freq_table = path.join(self.output_directory, self.OUTPUT_CLUSTER)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

                if self.annotate_ortholog:
                    matrix_generator = MatrixGenerator(MatrixGenerator.ORTHOLOG, ortholog_ids)
                    freq_table = path.join(self.output_directory, self.OUTPUT_ORTHOLOG)
                    matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.annotate_orthogroup:
                logging.warning(f"Not yet implemented")
                #self.annotate_orthogroup(genomes_list)

            if self.annotate_ko:
                annotation_type = AnnotationParser.BLASTPARSER
                logging.info('    - Annotating genomes with ko ids using DIAMOND')
                self.annotate_diamond(genomes_list, self.databases.KO_DB,
                                      annotation_type, AnnotationParser.KO,
                                      self.GENOME_KO)

                logging.info('    - Generating ko frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.KO)
                freq_table = path.join(self.output_directory, self.OUTPUT_KO)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.annotate_ko_hmm:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with ko ids using HMMs')
                self.hmmsearch_annotation(genomes_list,
                                          path.join(
                                              self.output_directory, self.GENOME_KO_HMM),
                                          self.databases.KO_HMM_DB,
                                          AnnotationParser.KO,
                                          annotation_type)

                logging.info('    - Generating ko frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.KO)
                freq_table = path.join(
                    self.output_directory, self.OUTPUT_KO_HMM)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.annotate_ec:
                annotation_type = AnnotationParser.BLASTPARSER
                logging.info('    - Annotating genomes with ec ids')
                self.annotate_diamond(genomes_list, self.databases.EC_DB, annotation_type,
                                      AnnotationParser.EC, self.GENOME_EC)

                logging.info('    - Generating ec frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.EC)
                freq_table = path.join(self.output_directory, self.OUTPUT_EC)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.annotate_pfam:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with pfam ids')
                self.hmmsearch_annotation(genomes_list,
                                          path.join(self.output_directory, self.GENOME_PFAM),
                                          self.databases.PFAM_DB,
                                          AnnotationParser.PFAM,
                                          annotation_type)

                logging.info('    - Generating pfam frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.PFAM)
                freq_table = path.join(self.output_directory, self.OUTPUT_PFAM)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.annotate_tigrfam:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with tigrfam ids')
                self.hmmsearch_annotation(genomes_list,
                                          path.join(self.output_directory, self.GENOME_TIGRFAM),
                                          self.databases.TIGRFAM_DB,
                                          AnnotationParser.TIGRFAM,
                                          annotation_type)

                logging.info('    - Generating tigrfam frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.TIGRFAM)
                freq_table = path.join(self.output_directory, self.OUTPUT_TIGRFAM)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if self.annotate_cazy:
                annotation_type = AnnotationParser.HMMPARSER
                logging.info('    - Annotating genomes with CAZY ids')
                self.hmmsearch_annotation(genomes_list,
                                          path.join(self.output_directory, self.GENOME_CAZY),
                                          self.databases.CAZY_DB,
                                          AnnotationParser.CAZY,
                                          annotation_type)

                logging.info('    - Generating CAZY frequency table')
                matrix_generator = MatrixGenerator(MatrixGenerator.CAZY)
                freq_table = path.join(self.output_directory, self.OUTPUT_CAZY)
                matrix_generator.write_matrix(genomes_list, self.count_domains, freq_table)

            if hasattr(list(genomes_list[0].sequences.values())[0], "prod_id"):
                logging.info('Generating .gff files:')
                self.generate_gff_files(genomes_list)

            logging.info('Renaming protein headers')
            self.rename_fasta(genomes_list)

            if not self.light:
                logging.info('Storing genome objects')
                self.pickle_objects(genomes_list)

            logging.info('Finished annotation')

        else:
            logging.error('No files found with %s suffix in input directory', self.suffix)
            raise Exception("No input files found")



#!/usr/bin/env python3
# pylint: disable=line-too-long
import random
import os
import logging
import multiprocessing as mp
from itertools import product, combinations, chain
from scipy import stats
import numpy as np
import statsmodels.sandbox.stats.multicomp as sm
from enrichm.draw_plots import Plot
from enrichm.databases import Databases
from enrichm.module_description_parser import ModuleDescription
from enrichm.parser import Parser, ParseAnnotate
from enrichm.writer import Writer
from enrichm.synteny_searcher import SyntenySearcher
################################################################################

def gene_fisher_calc(x):
    annotation, group_1, group_2 = x[0], x[1], x[2]

    dat = x[3:]

    if (dat[0][0]>0 or dat[1][0]>0) and (dat[0][1]>0 or dat[1][1]>0):
        score, pval = stats.fisher_exact(dat)
        if (dat[0][0] / sum(dat[0])) > (dat[1][0] / sum(dat[1])):
            enriched_in = group_1
        else:
            enriched_in = group_2
    else:
        enriched_in = 'NA'
        score, pval = 'nan', 1.0

    return [annotation, group_1, group_2, enriched_in] + dat[0] + dat[1] + [score, pval]

def mannwhitneyu_calc(x):
    # Mann Whitney U test
    annotation, group_1, group_2, group_1_module_annotations, group_2_module_annotations = x
    group_1_module_annotations = np.array(group_1_module_annotations[0])
    group_2_module_annotations = np.array(group_2_module_annotations[0])
    group_1_mean = np.mean(group_1_module_annotations)
    group_2_mean = np.mean(group_2_module_annotations)

    if(sum(group_1_module_annotations)>0 and sum(group_2_module_annotations)>0):

        if(len(set(group_1_module_annotations)) == 1
            or
           len(set(group_2_module_annotations)) == 1):
            mw_t_stat, mw_p_value = 'NA', 1
            enriched_in = 'NA'

        else:

            if group_1_mean > group_2_mean:
                enriched_in = group_1
            else:
                enriched_in = group_2

            mw_t_stat, mw_p_value = \
                stats.mannwhitneyu(group_1_module_annotations,
                                         group_2_module_annotations)
    else:
        mw_t_stat, mw_p_value = 'NA', 1
        enriched_in = 'NA'

    return [annotation, group_1, group_2, enriched_in, str(group_1_mean), str(group_2_mean), mw_t_stat, mw_p_value]

def zscore_calc(x):

    annotation, group_1, group_2, group_1_module_annotations, group_2_module_annotations = x
    group_1_module_annotations = group_1_module_annotations[0]
    group_2_module_annotations = group_2_module_annotations[0]

    if len(group_1_module_annotations)>1:
        reference = group_1_module_annotations
        reference_name = group_1
        genome = group_2_module_annotations[0]
        genome_name = group_2

    else:
        reference_name = group_2
        reference = group_2_module_annotations
        genome = group_1_module_annotations[0]
        genome_name = group_1

    if genome>0:
        reference_group_comp_sd = np.std(reference, axis=0)
        reference_group_comp_mean = np.mean(reference, axis=0)

        if (genome-reference_group_comp_mean)>0:

            if reference_group_comp_sd==0:
                z_score = np.inf
                p_value = 0.0
                enriched_in = 'NA'

            else:
                z_score = (genome-reference_group_comp_mean) / reference_group_comp_sd
                p_value = 2-2*stats.norm.cdf(z_score)
                enriched_in = genome_name

            return [annotation,
                    reference_name,
                    genome_name,
                    enriched_in,
                    str(reference_group_comp_mean),
                    str(reference_group_comp_sd),
                    str(genome),
                    str(z_score),
                    p_value]

################################################################################

class Enrichment:

    TIGRFAM = "tigrfam"
    PFAM = "pfam"
    KEGG = "kegg"
    CAZY = "cazy"
    EC = "ec"
    CLUSTER = "cluster"
    ORTHOLOG = "ortholog"
    OTHER = "other"

    def __init__(self):

        self.TIGRFAM_PREFIX = 'TIGR'
        self.PFAM_PREFIX = 'PF'
        self.KEGG_PREFIX = 'K'
        self.CAZY_PREFIX = ["GH", "AA", "GT", "PL", "CE", "CBM", "SLH", "dockerin", "cohesin", "GTCellulosesynt"]
        self.EC_PREFIX = ["1", "2", "3","4","5","6", "7"]
        self.PROPORTIONS = 'proportions.tsv'
        self.MODULE_COMPLETENESS = 'modules.tsv'

    def check_annotation_type(self, annotations):
        '''
        Takes a random sample of the rownames from the input matrix
        and based on the characters they start with, determines the
        type of annotation being used.

        Parameters
        ----------
        annotaitons     - List. A list of strings, each a rowname
                          from the original input matrix

        Output
        ------
        The annotation type
        '''

        sample = random.sample(annotations, 1)[0]
        cazy_prefix = ''

        for character in list(sample):

            if character.isdigit()!=True:

                if character!='_':
                    cazy_prefix+=character

        if sample.startswith(self.TIGRFAM_PREFIX):
            return self.TIGRFAM
        elif sample.startswith(self.KEGG_PREFIX):
            return self.KEGG
        elif sample.startswith(self.PFAM_PREFIX):
            return self.PFAM
        elif cazy_prefix in self.CAZY_PREFIX:
            return self.CAZY
        elif sample.split('.')[0] in self.EC_PREFIX:
            return self.EC
        else:
            return self.OTHER

    def weight_annotation_matrix(self,
                                 sample_abundance,
                                 annotation_abundance,
                                 sample_dict,
                                 annotations):

        output_dict = {sample_group: dict() for sample_group in sample_dict.keys()}

        logging.info('Aggregating abundances across samples')
        for group, samples in sample_dict.items():
            output_dict[group] = dict()

            for annotation in annotations:
                output_dict[group][annotation] = list()

                for sample in samples:
                    sample_annotation_abundance = 0.0

                    for genome, genome_annotation_dict in annotation_abundance.items():

                        if annotation in genome_annotation_dict:
                            value = genome_annotation_dict[annotation]
                            if genome in sample_abundance[sample]:
                                sample_annotation_abundance += sample_abundance[sample][genome]*value
                    output_dict[group][annotation].append(sample_annotation_abundance)

        return output_dict

    def calculate_portions(self,
                           annotations,
                           combination_dict,
                           annotations_dict,
                           genome_list,
                           proportions_cutoff):
        '''
        Calculates the portions of genome

        Parameters
        ----------
        modules             - List. List of all possible annotations for the given annotation
                              type (eg, all ko ids, or all pfam ids).
        combination_dict    - Dictionary. Metadata dictionary, with the groups as keys, and lists
                              of genomes as entries
        annotations_dict    - Dictionary. Annotation dictionary, with the the genome ids as keys,
                              and a list of annotations as the entry for each.
        genome_list         - List. List of strings, each one a genome name
        proportions_cutoff  - Float. Value with which to cutoff
        '''
        raw_proportions_output_lines        = [['Annotation'] + list(combination_dict.keys())]

        for annotation in annotations:

            annotation_values               = dict()
            raw_proportions_output_line = [annotation]

            for group_name, genome_list in combination_dict.items():
                if len(genome_list)>0:
                    coverage = len([genome for genome in genome_list
                                    if annotation in annotations_dict[genome]])
                    total    = float(len(genome_list))
                    entry    = coverage/total
                    annotation_values[group_name] = entry
                    raw_proportions_output_line.append(str(entry))
                else:
                    raw_proportions_output_line.append('0.0')

            raw_proportions_output_lines.append(raw_proportions_output_line)

        return raw_proportions_output_lines


    def get_gtdb_database_path(self, annotation_type, database):

        if annotation_type == self.KEGG:
            gtdb_annotation_matrix = database.GTDB_KO
        elif annotation_type == self.TIGRFAM:
            gtdb_annotation_matrix = database.GTDB_PFAM
        elif annotation_type == self.PFAM:
            gtdb_annotation_matrix = database.GTDB_TIGRFAM
        elif annotation_type == self.CAZY:
            gtdb_annotation_matrix = database.GTDB_CAZY
        elif annotation_type == self.EC:
            gtdb_annotation_matrix = database.GTDB_EC
        else:
            gtdb_annotation_matrix = None
        return gtdb_annotation_matrix

    def module_completeness(self, database, result_file_path, pval_cutoff):
        module_output = [["Module", "Lineage", "Total steps", "Steps covered", "Percentage covered", "Module description"]]
        module_descriptions = database.m()

        g1_sig_annotations = set()
        g2_sig_annotations = set()

        result_file_io = open(result_file_path)
        result_file_io.readline()

        for line in result_file_io:
            sline = line.strip().split('\t')
            if float(sline[-2])<pval_cutoff:
                if result_file_path.endswith("fisher.tsv"):
                    g1 = float(sline[4]) / (int(sline[4]) + int(sline[5]))
                    g2 = float(sline[6]) / (int(sline[6]) + int(sline[7]))
                elif result_file_path.endswith("cdf.tsv"):
                    g1 = float(sline[4])
                    g2 = float(sline[6])
                if g1>g2:
                    g1_sig_annotations.add(sline[0])
                else:
                    g2_sig_annotations.add(sline[0])

        for module, definition in database.m2def().items():

            if module not in database.signature_modules:
                pathway = ModuleDescription(definition)
                num_all = pathway.num_steps()
                g1_num_covered, _, _, _ = pathway.num_covered_steps(g1_sig_annotations)
                g1_perc_covered = g1_num_covered / float(num_all)

                g2_num_covered, _, _, _ = pathway.num_covered_steps(g2_sig_annotations)
                g2_perc_covered = g2_num_covered / float(num_all)

                if g1_perc_covered>0:
                    output_line = [module, sline[1], num_all, g1_num_covered, g1_perc_covered, module_descriptions[module]]
                    module_output.append(output_line)

                if g2_perc_covered>0:
                    output_line = [module, sline[2], num_all, g2_num_covered, g2_perc_covered, module_descriptions[module]]
                    module_output.append(output_line)

        prefix = '_vs_'.join([sline[1], sline[2]]).replace(' ', '_')

        return module_output, prefix

    def enrichment_pipeline(# Input options
           self, annotate_output, annotation_matrix, gff_files, metadata_path,
           abundances_path, abundance_metadata_path, transcriptome_path,
           transcriptome_metadata_path,
           # Runtime options
           pval_cutoff, proportions_cutoff,
           threshold, multi_test_correction, batchfile, processes, allow_negative_values,
           ko, pfam, tigrfam, cluster, ortholog, cazy, ec, ko_hmm, synteny_range, subblock_size,
           operon_mismatch_cutoff, operon_match_score_cutoff,
           # Output options
           output_directory):

        plot  = Plot()
        database = Databases()
        syntenysearcher = SyntenySearcher()

        if gff_files:
            logging.info("Parsing .gff file input(s)")
            annotations_dict = dict()
            gene_positions = dict()

            for gff_file in gff_files:
                position, counts = Parser.parse_gff(gff_file)
                annotations_dict.update(counts)
                gene_positions.update(position)

            annotations = set(chain(*[list(x.keys()) for x in annotations_dict.values()]))

        else:
            if annotate_output:
                logging.info('Parsing annotate output: %s' % (annotate_output))
                pa = ParseAnnotate(annotate_output, processes)

                if ko:
                    annotation_matrix = pa.ko
                elif ko_hmm:
                    annotation_matrix = pa.ko_hmm
                elif pfam:
                    annotation_matrix = pa.pfam
                elif tigrfam:
                    annotation_matrix = pa.tigrfam
                elif cluster:
                    annotation_matrix = pa.cluster
                elif ortholog:
                    annotation_matrix = pa.ortholog
                elif cazy:
                    annotation_matrix = pa.cazy
                elif ec:
                    annotation_matrix = pa.ec

            logging.info('Parsing annotation matrix')
            annotations_dict, _, annotations = Parser.parse_simple_matrix(annotation_matrix)

        annotation_type = self.check_annotation_type(annotations)

        if abundances_path:
            logging.info('Running abundances pipeline')
            logging.info('Parsing sample abundance')
            abundances_dict, _, _ = Parser.parse_simple_matrix(abundances_path)

            logging.info('Parsing sample metadata')
            _, _, ab_attribute_dict = Parser.parse_metadata_matrix(abundance_metadata_path)

            test = Test(annotations_dict,
                        None,
                        annotation_type,
                        threshold,
                        multi_test_correction,
                        processes,
                        database)

            weighted_abundance \
                = self.weight_annotation_matrix(abundances_dict,
                                                annotations_dict,
                                                ab_attribute_dict,
                                                annotations)
            results = test.test_weighted_abundances(weighted_abundance, annotations)

            for result in results:
                test_result_lines, test_result_output_file = result
                test_result_output_path = os.path.join(output_directory, test_result_output_file)
                Writer.write(test_result_lines, test_result_output_path)

        else:
            logging.info('Parsing metadata: %s' % metadata_path)
            metadata, metadata_value_lists, attribute_dict \
                = Parser.parse_metadata_matrix(metadata_path)
                
            if batchfile:
                gtdb_annotation_matrix = self.get_gtdb_database_path(annotation_type, database)

                batchfile_metadata, batchfile_metadata_value_lists, batchfile_attribute_dict = Parser.parse_metadata_matrix(batchfile)
                genomes_set = set(batchfile_metadata.keys())
                reference_genome_annotations, genomes_set = Parser.filter_large_matrix(genomes_set, gtdb_annotation_matrix)

                annotations_dict.update(reference_genome_annotations)
                new_batchfile_attribute_dict = dict()

                for group_name, accession_id_list in batchfile_attribute_dict.items():
                    filtered_accession_id_list = [accession_id for accession_id in accession_id_list if accession_id in genomes_set]

                    if len(filtered_accession_id_list)>0:
                        new_batchfile_attribute_dict[group_name] = filtered_accession_id_list

                attribute_dict.update(new_batchfile_attribute_dict)
                batchfile_metadata={group_name:batchfile_metadata[group_name] for group_name in genomes_set}
                metadata.update(batchfile_metadata)
                batchfile_metadata_value_lists = set(new_batchfile_attribute_dict.keys())
                metadata_value_lists = metadata_value_lists.union(batchfile_metadata_value_lists)

            logging.info("Comparing sets of genomes")
            combination_dict = dict()

            for combination in product(*list([metadata_value_lists])):
                genome_list = list()

                for genome, attributes in metadata.items():

                    for feature in combination:

                        if feature in attributes:
                            genome_list.append(genome)

                combination_dict['_'.join(combination)] = genome_list

            test = Test(annotations_dict, combination_dict, annotation_type, threshold, multi_test_correction, processes, database)
            results = test.test_pipeline(attribute_dict)

            for result in results:
                test_result_lines, test_result_output_file = result
                test_result_output_path = os.path.join(output_directory, test_result_output_file)
                Writer.write(test_result_lines, test_result_output_path)

        raw_proportions_output_lines = self.calculate_portions(annotations, combination_dict, annotations_dict, genome_list, proportions_cutoff)
        Writer.write(raw_proportions_output_lines, os.path.join(output_directory, self.PROPORTIONS))

        if gff_files:
            logging.info("Searching for co-located clusters of genes")
            synteny_results_output_lines, synteny_results_path \
                = syntenysearcher.search_for_blocks(results[0][0],
                                                    gene_positions,
                                                    metadata,
                                                    synteny_range,
                                                    subblock_size,
                                                    operon_mismatch_cutoff,
                                                    operon_match_score_cutoff)
            Writer.write(synteny_results_output_lines, os.path.join(output_directory, synteny_results_path))

        logging.info('Generating summary plots')
        if annotation_type == self.KEGG:
            logging.info('Finding module completeness in differentially abundant KOs')

            for result_file in os.listdir(output_directory):

                if(result_file.endswith("fisher.tsv") or result_file.endswith("cdf.tsv")):
                    plot.draw_barplots(os.path.join(output_directory, result_file), pval_cutoff, output_directory)
                    module_output, prefix = self.module_completeness(database, os.path.join(output_directory, result_file), pval_cutoff)
                    Writer.write(module_output, os.path.join(output_directory, prefix +'_'+ self.MODULE_COMPLETENESS))

        plot.draw_pca_plot(annotation_matrix, metadata_path, output_directory)

class Test(Enrichment):

    FISHER_HEADER = [['annotation', 'group_1', 'group_2', 'enriched_in', 'group_1_true', 'group_1_false',
                      'group_2_true', 'group_2_false', 'score', 'pvalue', 'corrected_pvalue', 'description']]

    MANNWHITNEYU_HEADER =[['annotation', 'group_1', 'group_2', 'enriched_in', 'group_1_mean', 'group_2_mean',
                           'score', 'pvalue', 'corrected_pvalue', 'description']]

    ZSCORE_HEADER = [['annotation', 'group_1', 'group_2', 'enriched_in', 'group_1_mean', 'group_1_sd',
                      'group_2_count', 'score', 'pvalue', 'corrected_pvalue', 'description']]

    PA                     = 'presence_absence'
    IVG_OUTPUT             = 'ivg_results.cdf.tsv'
    GENE_FISHER_OUTPUT     = 'gvg_results.fisher.tsv'
    GVG_OUTPUT             = 'gvg_results.mannwhitneyu.tsv'

    mtc_dict = {'b': 'Bonferroni',
                's': 'Sidak',
                'h': 'Holm',
                'hs': 'Holm-Sidak',
                'sh': 'Simes-Hochberg',
                'ho': 'Hommel',
                'fdr_bh': 'FDR Benjamini-Hochberg',
                'fdr_by': 'FDR Benjamini-Yekutieli',
                'fdr_tsbh': 'FDR 2-stage Benjamini-Hochberg',
                'fdr_tsbky': 'FDR 2-stage Benjamini-Krieger-Yekutieli',
                'fdr_gbs': 'FDR adaptive Gavrilov-Benjamini-Sarkar'}

    def __init__(self, genome_annotations, groups,
                 annotation_type, threshold, multi_test_correction,
                 processes, database):
        '''
        Collects functions to count and test differential abundance among groups of genomes.
        '''

        self.threshold              = threshold
        self.multi_test_correction  = multi_test_correction
        self.annotation_type        = annotation_type
        self.groups                 = groups
        self.pool                   = mp.Pool(processes = processes)
        self.m2def                  = database.m2def()
        self.m                      = database.m()
        self.clan2pfam              = database.clan2pfam()
        self.clan_to_description    = database.clan2name()
        self.k                      = database.k()
        self.tigrfamdescription     = database.tigrfamdescription()
        self.pfam2description       = database.pfam2description()
        self.ec2description         = database.ec2description()

        if annotation_type==self.PFAM:
            self.genome_annotations = dict()
            for key, item in genome_annotations.items():
                self.genome_annotations[key] = {key.split('.')[0]:entry for key,entry in item.items()}
        else:
            self.genome_annotations = genome_annotations

    def test_chooser(self, groups):
        groups = [len(x) for x in groups]

        # Enrichment
        if any(group == 1 for group in groups):
            enrichment_test = self.PA
            overrepresentation_test = stats.norm.cdf

        else:
            enrichment_test = stats.fisher_exact
            overrepresentation_test = stats.mannwhitneyu

        return enrichment_test, overrepresentation_test

    def correct_multi_test(self, pvalues):
        logging.info('Applying multi-test correction using the %s method' % (self.mtc_dict[self.multi_test_correction]) )
        corrected_pvals \
            = sm.multipletests(pvalues,
                               alpha        = self.threshold,
                               method       = self.multi_test_correction,
                               returnsorted = False,
                               is_sorted    = False)[1]

        return corrected_pvals

    def count(self, annotation, group, freq):

        if freq:
            group_true = list()

        else:
            group_true = 0

        group_false = 0

        for genome in self.groups[group]:

            if annotation in self.genome_annotations[genome]:
                if freq:
                    group_true.append(self.genome_annotations[genome][annotation])
                else:
                    if self.genome_annotations[genome][annotation]>0.0:
                        group_true+=1
                    else:
                        group_false+=1

            else:

                if freq:
                    group_true.append(0)

                else:
                    group_false+=1

        return group_true, group_false

    def gene_frequencies(self, group_1, group_2, freq=False):

        res_list    = list()
        annotations = set(chain(*self.genome_annotations.values()))

        for annotation in annotations:
            passed = True
            group_1_true, group_1_false \
                = self.count(annotation,
                              group_1,
                              freq)
            group_2_true, group_2_false \
                = self.count(annotation,
                              group_2,
                              freq)

            if freq:
                if(len([x for x in group_1_true if x!='0'])==0 and
                    len([x for x in group_2_true if x!='0'])==0 ):
                    passed = False
            else:
                if(group_1_true==0 and group_2_true==0):
                    passed = False

            if passed:
                res_list.append([annotation, group_1, group_2, [group_1_true, group_1_false], [group_2_true, group_2_false]])

        return res_list

    def corrected_pvals(self, output_lines):
        pvalues = [output_line[-1] for output_line in output_lines]
        corrected_pvalues = self.correct_multi_test(pvalues)

        return corrected_pvalues

    def add_descriptions(self, output_lines):
        if self.annotation_type == self.KEGG:
            desc = self.k

        if self.annotation_type == self.CAZY:
            desc = None

        if self.annotation_type == self.TIGRFAM:
            desc = self.tigrfamdescription

        if self.annotation_type == self.PFAM:
            desc = self.pfam2description

        if self.annotation_type == self.EC:
            desc = self.ec2description

        if self.annotation_type == self.OTHER:
            desc = None

        for line in output_lines:
            annotation = line[0]

            if desc:

                if annotation in desc:
                    line.append(desc[annotation])

                else:
                    line.append("NA")

            else:
                line.append("NA")

        return output_lines

    def test_weighted_abundances(self,
                                 weighted_abundance,
                                 annotations):

        logging.info('Calculating enrichment across samples using Mann-Whitney U test')
        results = list()

        for combination in combinations(weighted_abundance, 2):
            prefix = '_vs_'.join(
                [sorted(combination)[0], sorted(combination)[1]]).replace(' ', '_')
            res_list = list()

            for annotation in annotations:
                group_1 = weighted_abundance[combination[0]][annotation]
                group_2 = weighted_abundance[combination[1]][annotation]
                gene_count = [annotation, combination[0],
                              combination[1], [group_1], [group_2]]
                res_list.append(gene_count)

            output_lines = self.pool.map(mannwhitneyu_calc, res_list)

            for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                output_lines[idx].append(str(corrected_pval))
            output_lines = self.add_descriptions(output_lines)
            output_lines = self.MANNWHITNEYU_HEADER + output_lines
            results.append([output_lines, prefix + '_' + self.GVG_OUTPUT])

        return results

    def test_pipeline(self, group_dict):
        results = list()

        for combination in combinations(group_dict, 2):
            enrichment_test, overrepresentation_test = self.test_chooser( [group_dict[member] for member in combination] )
            prefix = '_vs_'.join([sorted(combination)[0], sorted(combination)[1]]).replace(' ', '_')

            logging.info('Comparing gene frequency among groups: %s', ', '.join(combination))

            if enrichment_test == stats.fisher_exact:
                logging.info('Testing gene enrichment using Fisher\'s exact test')
                gene_count = self.gene_frequencies(*combination)
                output_lines = self.pool.map(gene_fisher_calc, gene_count)

                for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                    output_lines[idx].append(str(corrected_pval))

                header = self.FISHER_HEADER
                output = self.GENE_FISHER_OUTPUT
                output_lines = self.add_descriptions(output_lines)
                output_lines = header + output_lines
                results.append([output_lines, prefix + '_' + output])

            elif enrichment_test == self.PA:
                logging.info('enrichment statistics not possible with only one genome to compare')
                logging.info('See prevalence matrix for unique genes in groups')

            logging.info('Comparing gene over-representation among genomes')

            if(overrepresentation_test == stats.mannwhitneyu):
                logging.info('Testing over-representation using Mann-Whitney U test')
                gene_count = self.gene_frequencies(*combination, True)
                output_lines = self.pool.map(mannwhitneyu_calc, gene_count)

                for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                    output_lines[idx].append(str(corrected_pval))

                header = self.MANNWHITNEYU_HEADER
                output = self.GVG_OUTPUT

            elif overrepresentation_test == stats.norm.cdf:
                logging.info('Testing over-representation using Z score test')
                gene_count = self.gene_frequencies(*combination, True)
                output_lines = self.pool.map(zscore_calc, gene_count)
                output_lines = [x for x in output_lines if x]

                for idx, corrected_pval in enumerate(self.corrected_pvals(output_lines)):
                    output_lines[idx].append(str(corrected_pval))

                header = self.ZSCORE_HEADER
                output = self.IVG_OUTPUT

            output_lines = self.add_descriptions(output_lines)
            output_lines = header + output_lines
            results.append([output_lines, prefix +'_'+ output])

        return results



#!/usr/bin/env python
# Imports
import unittest
import filecmp
import tempfile
import os
import sys

PATH_TO_DATA = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data')
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')] + sys.path

from enrichm.network_analyzer import NetworkAnalyser

class Tests(unittest.TestCase):

    abundance_metadata = "abundances.metadata.tsv"
    abundance = "abundances.tsv"
    genome_annotations = "genome.annotations.tsv"
    genome_metadata = "genome.metadata.tsv"
    queries_file = "queries.txt"
    network_data = "network_data"
    expected_explore_output = "explore_test"
    expected_pathway_output = "pathway_test"

    abundance_metadata_path = os.path.join(PATH_TO_DATA, network_data, abundance_metadata)
    abundance_path = os.path.join(PATH_TO_DATA, network_data, abundance)
    expected_explore_output_path = os.path.join(PATH_TO_DATA, network_data, expected_explore_output)
    genome_annotations_path = os.path.join(PATH_TO_DATA, network_data, genome_annotations)
    genome_metadata_path = os.path.join(PATH_TO_DATA, network_data, genome_metadata)
    expected_pathway_output_path = os.path.join(PATH_TO_DATA, network_data, expected_pathway_output)
    queries_file_path = os.path.join(PATH_TO_DATA, network_data, queries_file)

    def test_hello_pathway(self):
        '''
        Testing normal pathway workflow, no abundance or transcriptome or metabolome data included
        '''

        tmp = tempfile.mkdtemp()

        network_analyser = NetworkAnalyser()
        network_analyser.network_pipeline(NetworkAnalyser.PATHWAY,
                                          self.genome_annotations_path, self.genome_metadata_path,
                                          None, None, # No transcriptome data
                                          None, None, # No metagenome data
                                          None, # No metabolome data
                                          None, # No fisher results
                                          None, # No depth (explore runs only)
                                          list(), # Dont filter anything out
                                          ["map00860"], # Filter to keep it small.
                                          None, # No queries
                                          tmp # Output directory
                                          )
        

        expected_files = sorted(os.listdir(tmp))
        observed_files = sorted(os.listdir(self.expected_pathway_output_path))

        self.assertEqual(len(expected_files), len(observed_files))

        for expected_file, observed_file in zip(expected_files, observed_files):
            expected_file_path = os.path.join(tmp, expected_file)
            observed_file_path = os.path.join(self.expected_pathway_output_path, observed_file)
            # Are all files present?
            self.assertEqual(expected_file, observed_file)

            # Do the all look the same on the inside?
            with open(expected_file_path) as expected_file_io:

                with open(observed_file_path) as observed_file_io:
                    expected_lines = sorted(expected_file_io.readlines())
                    observed_lines = sorted(observed_file_io.readlines())
                    self.assertEqual(len(expected_lines), len(observed_lines))

                    for expected_line, observed_line in zip(expected_lines, observed_lines):
                        self.assertEqual(expected_line, observed_line)

    def test_hello_query(self):
        '''
        Testing normal explore workflow, no abundance or transcriptome or metabolome data included
        '''

        tmp = tempfile.mkdtemp()

        network_analyser = NetworkAnalyser()
        network_analyser.network_pipeline(NetworkAnalyser.EXPLORE,
                                          self.genome_annotations_path, self.genome_metadata_path,
                                          None, None, # No transcriptome data
                                          None, None, # No metagenome data
                                          None, # No metabolome data
                                          None, # No fisher results
                                          1, # No depth (explore runs only)
                                          list(), # Dont filter anything out
                                          list(), # Filter to keep it small.
                                          self.queries_file_path, # No queries
                                          tmp # Output directory
                                          )

        expected_files = sorted(os.listdir(tmp))
        observed_files = sorted(os.listdir(self.expected_explore_output_path))

        self.assertEqual(len(expected_files), len(observed_files))

        for expected_file, observed_file in zip(expected_files, observed_files):
            expected_file_path = os.path.join(tmp, expected_file)
            observed_file_path = os.path.join(self.expected_explore_output_path, observed_file)
            # Are all files present?
            self.assertEqual(expected_file, observed_file)

            # Do the all look the same on the inside?
            with open(expected_file_path) as expected_file_io:

                with open(observed_file_path) as observed_file_io:
                    expected_lines = sorted(expected_file_io.readlines())
                    observed_lines = sorted(observed_file_io.readlines())
                    self.assertEqual(len(expected_lines), len(observed_lines))

                    for expected_line, observed_line in zip(expected_lines, observed_lines):
                        self.assertEqual(expected_line, observed_line)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import logging
import os
import numpy as np
from enrichm.parser import ParseGenerate
from enrichm.writer import Writer
from enrichm.parser import Parser

class Predict:

    def __init__(self):
        self.predictions_output_file = 'predictions.tsv'
        self.predictions_header = ["Sample", "Prediction", "Probability"]

    def make_predictions(self, model, sample_list, content_list, attribute_dictionary):
        '''
        Inputs
        ------

        Outputs
        -------

        '''

        sample_list = np.array(sample_list)
        content_list = np.array(content_list)
        predictions = model.predict(content_list)
        probabilities = model.predict_proba(content_list)

        output_list = [self.predictions_header]

        for sample, prediction, probability in zip(sample_list, predictions, probabilities):
            max_prob = str(round(max(list(probability)), 2))
            prediction = str(attribute_dictionary[prediction])
            output_list.append([sample, prediction, max_prob])

        return output_list

    def predict_pipeline(self, forester_model_directory, input_matrix_path, output_directory):
        '''
        Inputs
        ------

        Outputs
        -------

        '''
        forester_model = ParseGenerate(forester_model_directory)

        logging.info('Parsing input')
        logging.info('Loading model: %s' % (forester_model.rf_model))

        logging.info('Parsing data')
        features, _, _ = Parser.parse_simple_matrix(input_matrix_path)

        sample_list = list()
        content_list = list()

        for sample, content in features.items():
            sample_list.append(sample)
            sample_content = list()

            for attribute in forester_model.attributes:

                if attribute in content:
                    sample_content.append(content[attribute])
                else:
                    sample_content.append('0')

            content_list.append(sample_content)

        logging.info('Making predictions')
        output_lines = self.make_predictions(forester_model.model,
                                             sample_list,
                                             content_list,
                                             forester_model.labels)

        Writer.write(output_lines, os.path.join(output_directory, self.predictions_output_file))



#!/usr/bin/env python3
from enrichm.sequence_io import SequenceIO
import logging
import os

###############################################################################

class Genome:
    '''
    A genome object which collects all the attributes of an input genome,
    including protein sequences and their annotations
    '''
    def __init__(self, light, path, nucl, gene, gff=False):
        seqio = SequenceIO()
        self.clusters = set()
        self.orthologs = set()
        self.protein_ordered_dict = dict()
        self.sequences = dict()
        self.cluster_dict = dict()
        self.ortholog_dict = dict()
        self.path = path
        self.gene = gene
        self.name = os.path.split(os.path.splitext(path)[0])[1]

        if light == False:

            if nucl is not None:
                self.nucl = nucl
                self.length = 0
                gc_list = 0.0

                for description, sequence in seqio.each(open(nucl)):
                    self.length += len(str(sequence))
                    gc_list 	+= (str(sequence).count('G') + str(sequence).count('C'))

                self.gc = round((gc_list/float(self.length))*100, 2)

            if gene:

                for protein_count, (protein_description, protein_sequence) in enumerate(seqio.each(open(path))):

                    for _, gene_sequence in seqio.each(open(gene)):
                        name = protein_description.partition(' ')[0]
                        sequence = Sequence(protein_description, protein_sequence, gene_sequence)
                        self.sequences[name] = sequence
                        self.protein_ordered_dict[protein_count] = name
            else:

                for protein_count, (protein_description, protein_sequence) in enumerate(seqio.each(open(path))):
                    name = protein_description.partition(' ')[0]
                    sequence = Sequence(protein_description, protein_sequence)
                    self.sequences[name] = sequence
                    self.protein_ordered_dict[protein_count] = name

        else:

            for protein_count, (description, _) in enumerate(seqio.each(open(path))):
                name = description.partition(' ')[0]
                sequence = Sequence(description)
                self.sequences[name] = sequence
                self.protein_ordered_dict[protein_count] = name

    def add(self, annotations, evalue_cutoff, bitscore_cutoff, percent_aln_query_cutoff,
            percent_aln_reference_cutoff, specific_cutoffs, annotation_type, ref_ids, pfam2clan=None):
        '''
        Adds a series of annotations to the proteins within a genome.

        Parameters
        ----------
        annotations						- String. Path to file containing either blast or
                                          domtblout hmmsearch, or mmseqs cluster results.
        evalue_cutoff					- Float. E-value threshold for annotations.
        bitscore_cutoff					- Float. Bit score threshold for annotations.
        percent_aln_query_cutoff		- Float. Threshold for the percent of the query
                                          that must be aligned to consider the annotation.
        percent_aln_reference_cutoff	- Float. Threshold for the percent of the reference
                                          that must be aligned to consider the annotation.
        annotation_type					- String. Either 'KO', 'TIGRFAM', 'PFAM',
                                          'HYPOTHETICAL' or 'COG'
        '''
        # Load up annotation parser, and tell it what annotation type to expect
        ap = AnnotationParser(annotation_type)

        # If annotation type is a hmmsearch result
        if(annotation_type == AnnotationParser.HMMPARSER):
            # Set up an iterator to produce the results
            logging.debug("    - Parsing hmmsearch chunk")

            if(ref_ids == AnnotationParser.PFAM or
               ref_ids == AnnotationParser.KO_HMM):
                percent_aln_query_cutoff = 0.0
                percent_aln_reference_cutoff = 0.0

            iterator = ap.from_hmmsearch_results(annotations, evalue_cutoff,
                                                 bitscore_cutoff, percent_aln_query_cutoff,
                                                 percent_aln_reference_cutoff,
                                        specific_cutoffs,
                                        (True if ref_ids == AnnotationParser.KO_HMM else False))

            if ref_ids == AnnotationParser.PFAM:
                self.pfam_dict = dict()
                refdict = self.pfam_dict

            elif ref_ids == AnnotationParser.KO_HMM:
                self.ko_dict = dict()
                refdict = self.ko_dict

            elif ref_ids == AnnotationParser.TIGRFAM:
                self.tigrfam_dict = dict()
                refdict = self.tigrfam_dict

            elif ref_ids == AnnotationParser.CAZY:
                self.cazy_dict = dict()
                refdict = self.cazy_dict

        # If annotation type is a blast result
        elif(annotation_type == AnnotationParser.BLASTPARSER):
            # Set up an iterator to produce the results
            logging.debug("    - Parsing blast chunk")
            iterator = ap.from_blast_results(annotations, evalue_cutoff,
                                             bitscore_cutoff, percent_aln_query_cutoff)

            if ref_ids == AnnotationParser.KO:
                self.ko_dict = dict()
                refdict = self.ko_dict

            elif ref_ids == AnnotationParser.EC:
                self.ec_dict = dict()
                refdict = self.ec_dict

        for seqname, annotations, evalue, annotation_range in iterator:
            self.sequences[seqname].add(annotations, evalue, annotation_range, ref_ids,
                                        pfam2clan=pfam2clan)

            for annotation in annotations:

                if annotation in refdict:
                    refdict[annotation].append(seqname)
                else:
                    refdict[annotation]=[seqname]

    def count(self, annotation, type):
        '''

        Parameters
        ----------
        annotation - String. An annotation ID to return a frequency for

        Output
        ------
        The number of times this annotation was encountered in the genome
        '''

        if type==AnnotationParser.HYPOTHETICAL:
            reference_dict = self.cluster_dict

        elif type==AnnotationParser.KO:
            reference_dict = self.ko_dict

        elif type==AnnotationParser.EC:
            reference_dict = self.ec_dict

        elif type==AnnotationParser.PFAM:
            reference_dict = self.pfam_dict

        elif type==AnnotationParser.TIGRFAM:
            reference_dict = self.tigrfam_dict

        if annotation in reference_dict:
            return len(reference_dict[annotation])

        else:
            return 0

    def ordered_sequences(self):
        '''
        Iterator that yields that all protein coding Sequence objects in a genome in order.
        '''
        for sequence_id in sorted(self.protein_ordered_dict.keys()):

            yield self.sequences[self.protein_ordered_dict[sequence_id]]

    def add_cluster(self, sequence_id, cluster_id):
        '''
        Add hypothetical clusters to genome sequences

        Inputs
        ------
        cluster_list - array. list where each entry is a list
                       of n = 2: sequence_name, cluster_name
        '''

        if cluster_id in self.cluster_dict:
            self.cluster_dict[cluster_id].append(sequence_id)
        else:
            self.cluster_dict[cluster_id] = [sequence_id]

        annotation = Annotation(cluster_id, 0, [-1,0], AnnotationParser.HYPOTHETICAL)

        self.sequences[sequence_id].cluster = cluster_id
        self.sequences[sequence_id].annotations.append(annotation)

        self.clusters.add(cluster_id)

    def add_ortholog(self, sequence_id, ortholog_id):
        '''
        Add hypothetical ortholog to genome sequences

        Inputs
        ------
        cluster_list - array. list where each entry is a list
                       of n = 2: sequence_name, cluster_name
        '''

        if ortholog_id in self.ortholog_dict:
            self.ortholog_dict[ortholog_id].append(sequence_id)
        else:
            self.ortholog_dict[ortholog_id] = [sequence_id]

        annotation = Annotation(ortholog_id, 0, [-1,0], AnnotationParser.ORTHOLOG)

        self.sequences[sequence_id].ortholog = ortholog_id
        self.sequences[sequence_id].annotations.append(annotation)
        self.orthologs.add(ortholog_id)

class Sequence(Genome):
    '''
    Sequence object which collects all attributes of a sequence including its length,
    and annotations. Can compare current annotation with new annotaitons.
    '''
    def __init__(self, description, sequence=None, gene=None):
        self.annotations = list()
        line_split = description.split(' # ')

        if gene:
            self.gene = gene
        if sequence:
            self.seq = str(sequence)
            self.length = int(len(sequence))

        if len(line_split)==5:
            self.seqname, self.startpos, self.finishpos, self.direction, stats \
                                = line_split
            self.prod_id, self.partial, self.starttype, self.rbs_motif, self.rbs_spacer,  self.gc \
                                = [x.split('=')[1] for x in stats.split(';')]

        elif len(line_split)==1:
            self.seqname = description
        else:
            raise Exception("Malformatted sequence file!")

    def all_annotations(self):
        '''
        Returns a list of all annotations assigned to this sequence
        '''
        result = []

        for annotation in self.annotations:
            result.append(annotation.annotation)

        return result

    def seqdict(self):
        '''
        Output
        ------
        A dictionary where each entry is a position in the sequence,
        and values are the annotation at that position. This is important
        in particular for pfam annotations as a protein can have >1
        domains.
        '''
        seq_dict = {x:None for x in range(self.length)}

        for annotation in self.annotations:

            for position in annotation.region:
                seq_dict[position] = annotation.annotation

        return seq_dict

    def what(self, query_region):
        '''
        Return annotations assigned to a list of positions within a sequence.

        Parameters
        ----------
        region 		- list. List of integers specifying the positions in the
                      sequence to return annotations for
        Outputs
        -------
        Returns a list of equal length to region, containing the annotation of
        each position.
        '''
        result = list()

        # Build reference dictionary for sequence
        seq_dict = self.seqdict()

        # Find annotation for each position in the sequence
        for position in query_region:
            result.append(seq_dict[position])

        return result

    def add(self, annotations, evalue, region, annotation_type, pfam2clan=None):
        '''
        Return annotations assigned to a list of positions within a sequence.

        Parameters
        ----------
        annotation 	- string. Annotation to assign to the sequence region
        evalue 		- float. Evalue awarded for the given annotation
        region		- list. List of integers specifying the positions in the
                      sequence to annotate
        '''

        new_annotations = [Annotation(annotation, evalue, region, annotation_type) for annotation in annotations]
        annotation_list = [annotation for annotation in self.annotations if annotation.type == new_annotations[0].type]

        if len(annotation_list) > 0:

            to_remove 	= list()
            to_check 	= annotation_list

            for new_annotation in new_annotations:
                overlap 	= [previous_annotation for previous_annotation in to_check
                               if len(previous_annotation.region.intersection(new_annotation.region)) > 0]

                if len(overlap)>0:

                    for overlapping_previous_annotation in overlap:

                        if annotation_type==AnnotationParser.PFAM:
                            if (new_annotation.annotation.split('.')[0] in pfam2clan
                            and overlapping_previous_annotation.annotation.split('.')[0] in pfam2clan
                            and pfam2clan[new_annotation.annotation.split('.')[0]]
                                == pfam2clan[overlapping_previous_annotation.annotation.split('.')[0]]
                            ):
                                if new_annotation.compare(overlapping_previous_annotation):
                                    to_remove.append(overlapping_previous_annotation)
                            else:
                                self.annotations.append(new_annotation)

                        else:
                            if new_annotation.compare(overlapping_previous_annotation):
                                to_remove.append(overlapping_previous_annotation)

                    if len(to_remove)>0:
                        self.annotations = [annotation for annotation in self.annotations
                                            if annotation not in to_remove]
                        self.annotations.append(new_annotation)

                else:
                        self.annotations.append(new_annotation)

        else:

            for new_annotation in new_annotations:
                self.annotations.append(new_annotation)

class Annotation(Sequence):
    '''
    Annotation object that collects all attributes associated with a given
    annotation, like where it is in the sequence, its e-value and bit score,
    and what type of annotation it is.
    '''
    def __init__(self, annotation, evalue, region, annotation_type):
        self.annotation = annotation
        self.evalue 	= float(evalue)
        self.region		= set(region)
        self.type   	= annotation_type

    def compare(self, other_annotation):
        '''
        Compares evalue of current annotation with the evalue of another annotation object.

        Parameters
        ----------
        other_annotation: Annotation object.

        Output
        ------
        Returns True if self is the better annotation or False if it isn't
        '''
        if self.evalue < other_annotation.evalue:
            return True

        else:
            return False

class AnnotationParser:
    '''
    Annotation parser class contains functions to parse hmmsearch domtblout and blast results
    currently for: KO, PFAM and TIGRFAM. COG to come
    '''
    KO      		= 'KO_IDS.txt'
    KO_HMM 			= 'KO_IDS.txt'
    EC				= 'EC_IDS.txt'
    PFAM    		= 'PFAM_IDS.txt'
    TIGRFAM 		= 'TIGRFAM_IDS.txt'
    CAZY      		= 'CAZY_IDS.txt'
    HYPOTHETICAL 	= 'HYPOTHETICAL.txt'
    ORTHOLOG 		= 'ORTHOLOG.txt'
    HMMPARSER 		= 'hmm'
    BLASTPARSER 	= 'blast'

    def __init__(self, annotation_type):
        pass

    def from_blast_results(self,
                           blast_output,
                           evalue_cutoff,
                           bitscore_cutoff,
                           percent_id_cutoff):
        '''
        Parse blast output in tab format.

        Parameters
        ----------
        blast_output 		- String. Path to blast output file containing results.
                              Must be in blast output format 6
        evalue_cutoff		- Float. E-value threshold for annotations.
        bitscore_cutoff		- Float. Bit score threshold for annotations.
        percent_id_cutoff 	- Float. Percent identity threshold for annotations.

        Yields
        ------
        A sequence name, annotation, E-value and region hit for every annotation result in
        blast_output that pass a series of specified cutoffs
        '''

        for sline in blast_output:
            # Parse out important information from each line in blast output
            evalue   = sline[10]
            bit      = sline[11]
            perc_id  = sline[2]
            seq_list = [int(x) for x in sline[6:8]]

            # If the annotation passes the specified cutoffs
            if(float(evalue) <= evalue_cutoff and
                float(bit) >= bitscore_cutoff and
                float(perc_id) >= percent_id_cutoff):
                    seqname = sline[0].split('~')[1]
                    annotation = sline[1].split('~')[1].split('+')

                    yield seqname, annotation, evalue, range(min(seq_list), max(seq_list))

    def from_hmmsearch_results(self,
                               hmmsearch_output_path,
                               evalue_cutoff,
                               bitscore_cutoff,
                               percent_aln_query_cutoff,
                               percent_aln_reference_cutoff,
                                specific_cutoffs,
                               acc = False):
        '''
        Parse input hmmsearch file

        Parameters
        ----------
        hmmsearch_output_path           - String. Path to domtblout file containing
                                          hmmsearch results.
        evalue_cutoff                   - Float. E-value threshold for annotations.
        bitscore_cutoff                 - Float. Bit score threshold for annotations.
        percent_aln_query_cutoff        - Float. Threshold for the percent of the query
                                          that must be aligned to consider the annotation.
        percent_aln_reference_cutoff    - Float. Threshold for the percent of the reference
                                          that must be aligned to consider the annotation.
        Yields
        ------
        A sequence name, accession, E-value and region hit for every annotation result in
        blast_output_path that pass a set of cutoffs
        '''

        # Filling in column
        for line in open(hmmsearch_output_path):

            # Skip headers
            if line.startswith('#'): continue

            # Parse HMMsearch line. '_'s represent unimportant entries. Line
            # is trimmed using [:22] to remove sequence description
            seqname, _, tlen, ko_hmm, accession, qlen, _, score, \
            _, _, _, _, i_evalue, dom_score, _, _, \
            _, seq_from, seq_to, _, _, _ = line.strip().split()[:22]

            # Determine sequence and HMM spans
            seq_list = [int(seq_from), int(seq_to)]

            # Calculate percent of the query and reference aligned to each-other.
            perc_seq_aln = (max(seq_list)-min(seq_list))/float(tlen)
            perc_hmm_aln = (max(seq_list)-min(seq_list))/float(qlen)

            if acc:
                accession = ko_hmm

            # If the annotation passes the specified cutoffs
            if specific_cutoffs:
                if ko_hmm in specific_cutoffs:
                    if specific_cutoffs[ko_hmm][1]=='full':
                        if float(score) < specific_cutoffs[ko_hmm][0]:
                            continue
                    elif specific_cutoffs[ko_hmm][1] == 'domain':
                        if float(dom_score) < specific_cutoffs[ko_hmm][0]:
                            continue

            if(float(i_evalue) <= evalue_cutoff and
                                    float(score) >= bitscore_cutoff and
                                    perc_seq_aln >= percent_aln_query_cutoff and
                                    perc_hmm_aln >= percent_aln_reference_cutoff):
                yield seqname, [accession], i_evalue, range(min(seq_list), max(seq_list))



#!/usr/bin/env python3
import os
import pickle
import json
import logging
import multiprocessing as mp
from enrichm.annotate import Annotate

################################################################################

def parse_genomes(path):
    genome = pickle.load(open(path, 'rb'))
    return genome

################################################################################

class Parser:
    '''
    A collection of functions to parse files in various formats.
    '''

    @staticmethod
    def parse_taxonomy(taxonomy_path):

        output_taxonomy_dictionary = dict()

        for line in open(taxonomy_path):
            genome, taxonomy_string = line.strip().split('\t')
            output_taxonomy_dictionary[genome] = taxonomy_string.split(';')

        return output_taxonomy_dictionary

    @staticmethod
    def parse_simple_matrix(matrix):
        with open(matrix) as matrix_io:
            colnames = matrix_io.readline().strip().split('\t')[1:]
            rownames = list()
            output_dict = {colname:dict() for colname in colnames}

            for line in matrix_io:
                sline = line.strip().split('\t')
                rowname, content = sline[0], sline[1:]

                if rowname not in rownames:
                    rownames.append(rowname)

                for key, value in zip(colnames, content):
                    output_dict[key][rowname] = float(value)

        return output_dict, colnames, rownames

    @staticmethod
    def parse_metadata_matrix(matrix_path):
        '''
        Parameters
        ----------
        matrix_path : String. Path to file containing a matrix of genome rownames.
        '''
        cols_to_rows = dict()
        nr_values = set()
        attribute_dict = dict()

        with open(matrix_path) as matrix_file_io:

            for line in matrix_file_io:
                rowname, entry = line.strip().split('\t')
                nr_values.add(entry)

                if entry in attribute_dict:
                    attribute_dict[entry].add(rowname)
                else:
                    attribute_dict[entry] = set([rowname])

                if rowname not in cols_to_rows:
                    cols_to_rows[rowname] = set([entry])
                else:
                    cols_to_rows[rowname].add(entry)

        return cols_to_rows, nr_values, attribute_dict

    @staticmethod
    def parse_single_column_text_file(text_file):
        entries = set()

        with open(text_file) as text_file_io:
            for entry in text_file_io:
                entries.add(entry.strip())

        return entries

    @staticmethod
    def filter_large_matrix(columns, matrix):
        '''
        description

        Inputs
        ------

        Outputs
        -------

        '''
        columns = list(columns)
        matrix_io = open(matrix)
        header = matrix_io.readline().strip().split('\t')

        indexes = list()
        include = list()

        for column in columns:

            if column in header:
                indexes.append(header.index(column))
                include.append(column)

        columns = include

        output_dict = {column:dict() for column in columns}

        for row in matrix_io:
            srow = row.strip().split()
            annotation = srow[0]

            for column, index in zip(columns, indexes):
                count = int(srow[index])

                if count > 0:
                    output_dict[column][annotation] = int(srow[index])

        return output_dict, columns

    @staticmethod
    def parse_gff(gff_file):
        feature_dict = dict()
        genome_to_annotations_dict = dict()
        
        gff_file_io = open(gff_file)
        
        for line in gff_file_io:

            if line.startswith('#'):
                continue
            contig, _, _, start_pos, finish_pos, _, strand, _, attributes = line.strip().split('\t')
            # Terminal ';' sometimes found in poorly formatted gffs. They gotta go.
            if attributes.endswith(';'):
                attributes = attributes[:-1]
            attributes_dict = dict()

            for attribute in attributes.split(';'):
                attribute_key, attribute_value = attribute.split('=')
                attribute_value_list = attribute_value.split(',')

                if attribute_key in attributes_dict:
                    raise Exception(f"Key duplicate in GFF file: {attribute_key}")
                else:
                    attributes_dict[attribute_key] = attribute_value_list 

            genome = attributes_dict['genome'][0]

            if genome not in genome_to_annotations_dict:
                genome_to_annotations_dict[genome] = dict()
                feature_dict[genome] = dict()

            for attribute in attributes_dict['annotations']:

                if contig not in feature_dict[genome]:
                    feature_dict[genome][contig] = dict()
                if attribute in feature_dict[genome][contig]:
                    feature_dict[genome][contig][attribute].append([int(start_pos), int(finish_pos), strand])
                    genome_to_annotations_dict[genome][attribute] += 1
                else:
                    feature_dict[genome][contig][attribute] = [[int(start_pos), int(finish_pos), strand]]
                    genome_to_annotations_dict[genome][attribute] = 1

        gff_file_io.close()

        return feature_dict, genome_to_annotations_dict

    @staticmethod
    def parse_tpm_values(tpm_values):

        from enrichm.databases import Databases

        k2r = Databases().k2r()

        output_dict = dict()
        annotation_types = set()
        genome_types = set()

        tpm_values_io = open(tpm_values, 'rb')
        tpm_values_io.readline()

        for line in tpm_values_io:
            gene, _, _, _, _, _, _, _, _, _, tpm, \
            _, _, annotation, sample = line.strip().split(b'\t')
            annotation_list = annotation.split(b',')
            tpm = float(tpm)
            genome = '_'.join(str(gene, "utf-8").split('_')[:2]) # temporary
            genome_types.add(genome)

            if sample not in output_dict:
                output_dict[sample] = dict()

            if genome not in output_dict[sample]:
                output_dict[sample][genome] = dict()

            for annotation_type in annotation_list:

                if str(annotation_type, "utf-8") in k2r:
                    reactions = k2r[str(annotation_type, "utf-8")]

                    for reaction in reactions:
                        reaction = str.encode(reaction)

                        if reaction not in output_dict[sample][genome]:
                            output_dict[sample][genome][reaction] = 0.0
                            annotation_types.add(reaction)

                        output_dict[sample][genome][reaction] += tpm
        return output_dict, annotation_types, genome_types

    @staticmethod
    def parse_enrichment_output(enrichment_output):
        fisher_results = dict()

        for file in os.listdir(enrichment_output):

            if file.endswith("fisher.tsv"):
                file = os.path.join(enrichment_output, file)
                file_io = open(file)
                file_io.readline()

                for line in file_io:
                    split_line = line.strip().split('\t')

                    if len(fisher_results) == 0:
                        fisher_results[split_line[1]] = list()
                        fisher_results[split_line[2]] = list()

                    if float(split_line[-2])<0.05:
                        g1_t = float(split_line[3])
                        g1_f = float(split_line[4])
                        g2_t = float(split_line[5])
                        g2_f = float(split_line[6])

                        if g1_t == 0:
                            fisher_results[split_line[2]].append( split_line[0] )

                        elif g2_t == 0:
                            fisher_results[split_line[1]].append( split_line[0] )

                        elif ( ((g1_t/(g1_t+g1_f))) / ((g2_t/(g2_t+g2_f))) )>1:
                            fisher_results[split_line[1]].append( split_line[0] )

                        else:
                            fisher_results[split_line[2]].append( split_line[0] )

        if len(fisher_results.keys())>0:
            return fisher_results

        else:
            raise Exception("Malformatted enrichment output: %s" % enrichment_output)


class ParseAnnotate:

    def __init__(self, enrichm_annotate_output, processes):
        self.path = enrichm_annotate_output
        # Parse genome objects
        self.genome_pickle_file_path \
            = os.path.join(enrichm_annotate_output, Annotate.GENOME_OBJ)
        self.processes \
            = processes

        ko = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_KO)
        if os.path.isfile(ko):
            self.ko = ko
        else:
            self.ko = None
        ko_hmm = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_KO_HMM)
        if os.path.isfile(ko_hmm):
            self.ko_hmm = ko_hmm
        else:
            self.ko_hmm = None
        pfam = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_PFAM)
        if os.path.isfile(pfam):
            self.pfam = pfam
        else:
            self.pfam = None
        tigrfam = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_TIGRFAM)
        if os.path.isfile(tigrfam):
            self.tigrfam = tigrfam
        else:
            self.tigrfam = None
        cazy = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_CAZY)
        if os.path.isfile(cazy):
            self.cazy = cazy
        else:
            self.cazy = None
        ec = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_EC)
        if os.path.isfile(ec):
            self.ec = ec
        else:
            self.ec = None
        cluster = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_CLUSTER)
        if os.path.isfile(cluster):
            self.cluster = cluster
        else:
            self.cluster = None
        ortholog = os.path.join(enrichm_annotate_output, Annotate.OUTPUT_ORTHOLOG)
        if os.path.isfile(ortholog):
            self.ortholog = ortholog
        else:
            self.ortholog = None

    def parse_pickles(self, path, genome_list):
        '''
        Opens the pickled genome objects from a previous run of enrichm
        annotate

        Parameters
        ----------
        enrichm_annotate_output - String. Output directory of a previous run
                                  of enrichm annotate (At lease version 0.0.7)
                                  or above
        Outputs
        -------
        List of Genome objects
        '''

        output_genome_list = list()
        paths = list()

        for pickled_genome in genome_list:
            pickled_genome_path = os.path.join(path, pickled_genome + '.pickle')
            if os.path.isfile(pickled_genome_path):
                paths.append(pickled_genome_path)

        self.pool = mp.Pool(processes=self.processes)
        output_genome_list = self.pool.map_async(parse_genomes, paths)
        output_genome_list.wait()
        genome_objects = output_genome_list.get()
        self.pool.close()

        return genome_objects

class ParseGenerate:

    def __init__(self, forester_model_directory):
        self.labels_dict = "labels_dict.pickle"
        self.rf_model = "rf_model.pickle"
        self.attribute_importances = "attribute_importances.tsv"
        self.forester_model_directory = forester_model_directory

        for content in os.listdir(forester_model_directory):
            content_path = os.path.join(forester_model_directory, content)

            if content == self.labels_dict:
                with open(content_path, 'rb') as content_path_io:
                    self.labels = pickle.load(content_path_io)
            elif content == self.rf_model:
                with open(content_path, 'rb') as content_path_io:
                    self.model = pickle.load(content_path_io)
            elif content == self.attribute_importances:
                self.attributes = list()
                with open(content_path) as content_path_io:
                    _ = content_path_io.readline() # Junk

                    for line in content_path_io:
                        attribute, _ = line.strip().split('\t')
                        self.attributes.append(attribute)

        if None in [self.labels, self.model, self.attributes]:
            raise Exception("Malformatted forester model directory: %s" \
                            % (forester_model_directory))

class RulesJson:
    _CURRENT = "0"
    VERSION = 'version'
    
    _GENERIC_KEYS = {
        "0": [
            "modules", # Modules with rules
            "version" # Version of database
        ]
    }
    _REQUIRED_KEYS = {"0": # DB version
                         # Type of rule
                        {"synteny": # Define synteny blocks
                            # values of rule
                            ["genes", # genes in synteny block
                             "range", # range of nucleic acids within which they are found
                             "breaks", # How many breaks are allowed in the synteny block. Ignored if strict.
                             "minsubblocksize", # How many genes must be found in split  synteny blocks. Ignored if strict.
                             "min"], # Minimum number of genes that must be found in synteny to pass.
                        #"homology": {},
                        "required": {}
                        }
                     }

    
    def load(self, rules_json_filepath):
        logging.debug(f"Loading rules JSON file: {rules_json_filepath}")

        rules_json_filepath_io = open(rules_json_filepath)
        self.contents_dict = json.load(rules_json_filepath_io)
        rules_json_filepath_io.close()

        logging.debug("Finding rules version")
        current_package_version = self.contents_dict[self.VERSION]

        logging.debug(f"Loading rules version {current_package_version}")
        if current_package_version == '0':
            rj = RulesJsonVersion0(self.contents_dict, rules_json_filepath, current_package_version)
        else:
            raise Exception(f"JSON rules file version invalid: {current_package_version}")

        logging.debug("Checking package format is valid")
        self.check_version_keys(self._GENERIC_KEYS[current_package_version], self.contents_dict)

        for module_rules in self.contents_dict['modules'].values():
            self.check_version_keys(self._REQUIRED_KEYS[current_package_version], module_rules)

        return rj

    def get_version(self):
        if self.VERSION in self.contents_dict:
            return self.contents_dict[self.VERSION]
        else:
            raise Exception("No version information in rules JSON file")

    def check_version_keys(self, version_keys, contents_dict):

        if isinstance(version_keys, dict):

            for key, item in version_keys.items():
                self.check_key(key, contents_dict, item)

        elif isinstance(version_keys, list):

            for key in version_keys:
                self.check_key(key, contents_dict)

    def check_key(self, key, contents, item=None):
        
        if isinstance(contents, dict):
            if key in contents:
                if item:
                    self.check_version_keys(item, contents[key])
            else:
                raise Exception(f"Malformatted rules file. Key not found: {key}")
        
        elif isinstance(contents, list):
            for content in contents:
                if key in content:
                    if item:
                        self.check_version_keys(item, contents[key])
                else:
                    raise Exception(f"Malformatted rules file. Key not found: {key}")

class RulesJsonVersion0(RulesJson):

    def __init__(self, _contents_json, _filename, _version):
        self.VERSION = 0
        self._contents_json = _contents_json
        self._filename = _filename
        self._version = _version
        self.members = set(self._contents_json['modules'].keys())

    def get_synteny_rules(self, module_name):
        return self._contents_json['modules'][module_name]['synteny']

    def get_homology_rules(self, module_name):
        return self._contents_json['modules'][module_name]['homology']

    def get_required_rules(self, module_name):
        return self._contents_json['modules'][module_name]['required']



#!/usr/bin/env python3
import logging
import os
import statistics
import itertools
from enrichm.network_builder import NetworkBuilder
from enrichm.databases import Databases
from enrichm.parser import Parser
from enrichm.writer import Writer
###############################################################################

class NetworkAnalyser:
    """
    Prepare metagenome, metatranscriptome, metabolomic data for constructing 
    SIF network files.
    """
    MATRIX          = 'matrix'
    NETWORK         = 'network'
    EXPLORE         = 'explore'
    DEGRADE         = 'degrade'
    PATHWAY         = 'pathway'
    ANNOTATE        = 'annotate'
    ENRICHMENT      = 'enrichment'
    MODULE_AB       = 'module_ab'
    TRAVERSE        = 'traverse'

    NETWORK_OUTPUT_FILE  = 'network.tsv'
    METADATA_OUTPUT_FILE = 'metadata.tsv'
    TRAVERSE_OUTPUT_FILE = 'traverse.tsv'

    def __init__(self):
        self.databases = Databases()
        self.reactions = self.databases.r()
        self.reaction_to_ko = self.databases.r2k()

    def average(self, input_dictionary):
        '''
        Take the average of the values of a dictionary of dictionaries
        '''
        
        for sample_group, group_dict in input_dictionary.items():

            for group, reaction_dict in group_dict.items():

                for reaction, value in reaction_dict.items():
                    input_dictionary[sample_group][group][reaction] = sum(value) / len(value)

        return input_dictionary

    def median_genome_abundance(self, sample_abundance_dict, sample_metadata):
        """
        Create a dictionary with the median abundance in sample_abundance_dict
        using sample_metadata as a reference.
        """
        
        median_sample_abundance = dict()

        for group, samples in sample_metadata.items():
            median_sample_abundance[group] = dict()
            sample_dictionaries = [sample_abundance_dict[sample] for sample in samples]
            genomes = set(list(itertools.chain(*[list(sample_dictionary.keys()) for sample_dictionary in sample_dictionaries])))

            for genome in genomes:
                abundances = [sample_dictionary[genome] for sample_dictionary in sample_dictionaries]
                median_sample_abundance[group][genome] = statistics.median(abundances)

        return median_sample_abundance

    def normalise_by_abundance(self, median_sample_abundances, reaction_abundance_dict, group_to_genome, genome_to_group, genome_groups):

        normalised_abundance_dict = dict()
        for sample_group in list(median_sample_abundances.keys()):
            normalised_abundance_dict[sample_group] = dict()

            for genome_group in genome_groups:
                normalised_abundance_dict[sample_group][genome_group] = dict()

        for sample_group, genome_abundances in median_sample_abundances.items():

            for genome, genome_abundance in genome_abundances.items():

                if(genome in genome_to_group and
                   genome in reaction_abundance_dict):

                    for reaction in list(reaction_abundance_dict[genome].keys()):

                        normalised_value = reaction_abundance_dict[genome][reaction]*genome_abundance

                        genome_group = next(iter(genome_to_group[genome]))

                        if reaction in normalised_abundance_dict[sample_group][genome_group]:
                            normalised_abundance_dict[sample_group][genome_group][reaction].append( normalised_value )
                        else:
                            normalised_abundance_dict[sample_group][genome_group][reaction] = [normalised_value]

        return normalised_abundance_dict

    def average_tpm_by_sample(self, tpm_results, sample_metadata):
        output_dict = dict()
        tpm_dict, annotations, genomes = tpm_results

        for group, samples in sample_metadata.items():
            output_dict[group] = dict()

            for sample in samples:

                for annotation in annotations:

                    if str.encode(sample) in tpm_dict:

                        for genome in genomes:

                            if genome not in output_dict[group]:
                                output_dict[group][genome] = dict()

                            if annotation not in output_dict[group][genome]:
                                output_dict[group][genome][annotation] = list()

                            if genome in tpm_dict[str.encode(sample)]:

                                if annotation in tpm_dict[str.encode(sample)][genome]:
                                    output_dict[group][genome][annotation].append(tpm_dict[str.encode(sample)][genome][annotation])
                                else:
                                    output_dict[group][genome][annotation].append(0.0)

                            else:
                                output_dict[group][genome][annotation].append(0.0)

            for genome, values in output_dict[group].items():

                for annotation in values:
                    output_dict[group][genome][annotation] = sum(output_dict[group][genome][annotation])/len(output_dict[group][genome][annotation])

        return output_dict

    def average_tpm_values(self, transriptome_abundance_dict, group_metadata):
        output_dict = dict()
        reactions = list(self.reactions.keys())
        
        for genome_group_name, group_reaction_abundance_dict in transriptome_abundance_dict.items():
            output_dict[genome_group_name] = dict()

            for group, members in group_metadata.items():
                output_dict[genome_group_name][group] = dict()

                for reaction in reactions:
                    to_average = list()

                    for member in members:

                        if member in group_reaction_abundance_dict:

                            if str.encode(reaction) in group_reaction_abundance_dict[member]:
                                to_average.append(group_reaction_abundance_dict[member][str.encode(reaction)])
                            else:
                                to_average.append(0.0)
                        else:
                            to_average.append(0.0)

                    average_value = sum(to_average) / len(to_average)
                    output_dict[genome_group_name][group][reaction] = average_value

        return output_dict

    def aggregate_dictionary(self, reference_dict, matrix_dict):

        output_dict_mean   = dict()

        for sample, ko_abundances in matrix_dict.items():
            output_dict_mean[sample]   = dict()

            for reaction, ko_list in reference_dict.items():
                abundances = list()

                for ko in ko_list:

                    if ko in ko_abundances:
                        if ko_abundances[ko]>0:
                            abundances.append(ko_abundances[ko])

                    else:
                        logging.debug("ID not found in input matrix: %s" % ko)

                if any(abundances):
                    abundance_mean = sum(abundances)/len(abundances) # average of the abundances...

                else:
                    abundance_mean = 0

                output_dict_mean[sample][reaction] = abundance_mean

        return output_dict_mean

    def mock_metadata(self, genomes):
        genome_to_group = {genome:set([genome]) for genome in genomes}
        genome_groups = set(genomes)
        group_to_genome = dict(genome_to_group) # Make a copy here
        return genome_to_group, genome_groups, group_to_genome

    def network_pipeline(self,
           subparser_name,
           matrix, genome_metadata_path,
           transcriptome_abundances_path, transcriptome_metadata_path,
           metagenome_abundances, metagenome_metadata_path,
           metabolome,
           enrichment_output,
           depth, filter, limit, queries, output_directory):
        '''
        Parameters
        ----------
        matrix
        transcriptome_abundances_path
        metagenome_abundances
        metagenome_metadata_path
        metabolome
        enrichment_output
        depth
        filter
        limit
        queries
        output_directory
        '''
        orthology_matrix, genome_names, _ = Parser.parse_simple_matrix(matrix)
        if genome_metadata_path:
            genome_to_group, genome_groups, group_to_genome = \
                    Parser.parse_metadata_matrix(genome_metadata_path)
        else:
            genome_to_group, genome_groups, group_to_genome = \
                    self.mock_metadata(genome_names)

        reaction_matrix = self.aggregate_dictionary(self.reaction_to_ko, orthology_matrix)

        # Read in fisher results
        if enrichment_output:
            logging.info('Parsing input enrichment results')
            fisher_results = Parser.parse_enrichment_output(enrichment_output)
        else:
            logging.info('No enrichment results provided')
            fisher_results = None

        # Read in metabolome abundances
        if metabolome:
            logging.info('Parsing metabolome abundances')
            abundances_metabolome = Parser.parse_simple_matrix(metabolome)
        else:
            logging.info('No metabolome abundances provided')
            abundances_metabolome = None

        # Read in genome metagenome_abundances
        if metagenome_abundances:
            logging.info('Parsing input genome abundances')
            sample_abundance = Parser.parse_simple_matrix(metagenome_abundances)[0]
            sample_metadata = Parser.parse_metadata_matrix(metagenome_metadata_path)[2]
        else:
            # FIXME : There's always a better way than faking it.
            logging.info('No genome abundances provided')
            sample_abundance = {'MOCK': {x:1 for x in list(reaction_matrix.keys())} }
            sample_metadata = {"abundance": ['MOCK']}

        median_sample_abundances = self.median_genome_abundance(sample_abundance, sample_metadata)
        normalised_abundance_dict = self.normalise_by_abundance(median_sample_abundances, reaction_matrix, group_to_genome, genome_to_group, genome_groups)
        abundances_metagenome = self.average(normalised_abundance_dict)

        # Read in expression (TPM) values
        if transcriptome_abundances_path:
            logging.info("Parsing detectM TPM abundances")
            transcriptome_metadata = Parser.parse_metadata_matrix(transcriptome_metadata_path)[2]
            transcriptome_abundance_dict = self.average_tpm_by_sample(Parser.parse_tpm_values(transcriptome_abundances_path), transcriptome_metadata)
            transcriptome_abundances = self.average_tpm_values(transcriptome_abundance_dict, group_to_genome)
        else:
            transcriptome_abundances = None

        network_builder = NetworkBuilder(group_to_genome, abundances_metagenome,
                                         transcriptome_abundances, abundances_metabolome, 
                                         fisher_results)

        # Run the subcommand specified
        if subparser_name == self.EXPLORE:
            network_lines, node_metadata = network_builder.query_matrix(queries, depth)
        elif subparser_name == self.PATHWAY:
            network_lines, node_metadata = network_builder.pathway_matrix(limit, filter)

        # Write the outputs
        Writer.write(network_lines, os.path.join(output_directory, self.NETWORK_OUTPUT_FILE))
        Writer.write(node_metadata, os.path.join(output_directory, self.METADATA_OUTPUT_FILE))
        
        logging.info('Finished the %s pipeline' % subparser_name)


#!/usr/bin/env python3
# Imports
import os
import urllib.request
import shutil
import logging
from pathlib import Path
from enrichm.toolbox import run_command
# Local
###############################################################################

class Data:
    '''
    Utilities for archiving, downloading and updating databases.
    '''
    db_var = "ENRICHM_DB"
    if db_var in os.environ:
        DATABASE_DIR 	= os.environ[db_var]
    else:
        DATA_PATH 		= str(Path.home())
        DATABASE_DIR	= os.path.join(DATA_PATH, 'databases')
    VERSION 		= 'VERSION'
    ARCHIVE_SUFFIX 	= '.tar.gz'

    def __init__(self):
        self.ftp = 'https://data.ace.uq.edu.au/public/enrichm/'

    def _archive_db(self, old_db_file):
        '''
        Archive an old database file

        Parameters
        ----------
        old_db_file	- String. File name of old database file to archive
        '''

        old_database_path = os.path.join(self.DATABASE_DIR, "old")
        if not os.path.isdir(old_database_path):
            logging.info('Creating directory to store databases: %s' % (old_database_path))
            os.makedirs(old_database_path)

        old_db_path_archive \
            = os.path.join(old_database_path, old_db_file + self.ARCHIVE_SUFFIX)
        old_db_path \
            = os.path.join(self.DATABASE_DIR, old_db_file)

        logging.info('Compressing old database')
        cmd = "tar -cvzf %s %s > /dev/null" % (old_db_path_archive, old_db_path)
        run_command(cmd)

        logging.info('Cleaning up')
        shutil.rmtree(old_db_path)

    def _download_db(self, new_db_file):
        '''
        Download and decompress a new database file

        Parameters
        ----------
        new_db_file	- String. File name of new database to download and decompress.
        '''

        new_db_path_archive \
            = os.path.join(self.DATABASE_DIR, new_db_file)
        
        logging.info('Downloading new database: %s', new_db_file)
        cmd = f'wget \
                    -q {self.ftp + new_db_file} \
                    -O {new_db_path_archive}'
        run_command(cmd)
        
        cmd = f'wget \
                    -q {self.ftp + self.VERSION} \
                    -O {os.path.join(self.DATABASE_DIR, self.VERSION)}'
        run_command(cmd)

        logging.info('Decompressing new database')
        cmd = 'tar -xvzf %s -C %s > /dev/null' % (new_db_path_archive, self.DATABASE_DIR)
        run_command(cmd)

        logging.info('Cleaning up')
        os.remove(new_db_path_archive)

    def do(self, uninstall, create, dry):
        '''
        Check database versions, if they're out of date, archive the old and download the new.
        '''

        if uninstall:

            for file in os.listdir(self.DATABASE_DIR):
                file_path = os.path.join(self.DATABASE_DIR, file)

                if os.path.isdir(file_path):
                    shutil.rmtree(file_path)
                else:
                    os.remove(file_path)

            os.rmdir(self.DATABASE_DIR)

        elif create:
            import IPython; IPython.embed()
            try:
                version_remote = urllib.request.urlopen(self.ftp + self.VERSION).readline().strip().decode("utf-8")
            except:
                raise Exception(
                    "Unable to find most current EnrichM database VERSION in ftp. Please complain at https://github.com/geronimp/enrichM")

            if os.path.isdir(self.DATABASE_DIR):
                version_local_path = os.path.join(self.DATABASE_DIR, self.VERSION)

                if os.path.isfile(version_local_path):
                    version_local = open(version_local_path).readline().strip()
                else:
                    raise Exception(
                        "Unable to locate enrichM database! Please specify its location by creating a local BASH variable called ENRICHM_DIR: export ENRICHM_DB=/path/to/database/")

                if version_local!=version_remote:
                    logging.info('New database found. Archiving old database.')
                    self._archive_db(version_local.replace(self.ARCHIVE_SUFFIX,''))
                    self._download_db(version_remote)
                else:
                    logging.info('Database is up to date!')

            else:
                logging.info('Creating folder to store databases.')
                os.makedirs(self.DATABASE_DIR)
                self._download_db(version_remote)




#!/usr/bin/env python3
from enrichm.writer import Writer
from enrichm.parser import Parser
from enrichm.databases import Databases
from enrichm.enrichment import mannwhitneyu_calc
from itertools import combinations

import logging
import os

class Uses:

    def __init__(self):
        
        databases = Databases()
        
        self.reaction_to_ko = databases.r2k()
        self.compound_to_reaction = databases.c2r()
        self.compounds = databases.c()

        self.positive = 'positive'
        self.negative = 'negative'

        self.abundace = "frequency_matrix.tsv"
        self.enrichment = "enrichment_results.tsv"
        self.abundace_header = ["Compound"]
        self.enrichment_header = ["Compound", "Group_1", "Group_2", "group_1_mean", "group_2_mean",
                                  "score", "pvalue", "description"]

    def gather_present_annotations(self, column_annotations):

        present_annotations = set()

        for annotation, copy_number in column_annotations.items():

            if copy_number>0:
                present_annotations.add(annotation)

        return present_annotations

    def uses(self, compound_list, annotations, column_names, count):
        output_lines_abundance = [self.abundace_header + column_names]
        enrichment_tallys = dict()

        for compound in compound_list:

            if compound in self.compound_to_reaction:

                enrichment_tallys[compound] = dict()
                abundance_line = [compound + '~' + self.compounds[compound]]

                for column_header in column_names:
                    column_positive_tally = 0
                    column_negative_tally = 0

                    # Gather all annotations present for this column (genome)
                    present_annotations = self.gather_present_annotations(annotations[column_header])
                    for reaction in self.compound_to_reaction[compound]:

                        # If there are more than 0 KOs that carry out the reaction present in
                        # the genome
                        if reaction in self.reaction_to_ko:
                            overlapping_annotations = present_annotations.intersection(self.reaction_to_ko[reaction])

                            if len(overlapping_annotations)>0:

                                if count:

                                    for annotation in overlapping_annotations:
                                        column_positive_tally+=annotations[column_header][annotation]

                                else:
                                    column_positive_tally+=1

                            else:
                                column_negative_tally+=1

                    abundance_line.append(column_positive_tally)
                    enrichment_tallys[compound][column_header] \
                        = {self.positive:column_positive_tally,
                           self.negative:column_negative_tally}

            output_lines_abundance.append(abundance_line)

        return output_lines_abundance, enrichment_tallys

    def enrichment(self, enrichment_tallys, metadata):
        output_lines = [self.enrichment_header]

        for compound, tallys in enrichment_tallys.items():

            for group_1_name, group_2_name in combinations(metadata.keys(), 2):
                group_1_tallys = {genome:tallys[genome] for genome in metadata[group_1_name]}
                group_2_tallys = {genome:tallys[genome] for genome in metadata[group_2_name]}
                group_1_values =  [tally[self.positive] for tally in group_1_tallys.values()]
                group_2_values =  [tally[self.positive] for tally in group_2_tallys.values()]

                output_line = mannwhitneyu_calc((compound, group_1_name, group_2_name, [group_1_values, None], [group_2_values, None]))
                output_lines.append(output_line + [self.compounds[compound]])

        return output_lines

    def uses_pipeline(self, compounds_list_path, annotation_matrix_path, metadata_path, output,
                      count):
        logging.info('Parsing input compounds list')
        compound_list = Parser.parse_single_column_text_file(compounds_list_path)
        logging.info('Parsing input annotations')
        annotations_dict, column_names, annotations = Parser.parse_simple_matrix(annotation_matrix_path)
        logging.info('Parsing input metadata')
        metadata, metadata_value_lists, attribute_dict = Parser.parse_metadata_matrix(metadata_path)
        logging.info('Tallying genes that use specified compounds')
        output_lines_abundance, enrichment_tallys = self.uses(compound_list, annotations_dict, column_names, count)
        logging.info('Writing file: %s' % self.abundace)
        Writer.write(output_lines_abundance, os.path.join(output, self.abundace))
        logging.info('Calculating enrichment between groups for each compound')
        output_lines_enrichment = self.enrichment(enrichment_tallys, attribute_dict)
        logging.info('Writing file: %s' % self.enrichment)
        Writer.write(output_lines_enrichment, os.path.join(output, self.enrichment))
        logging.info('Finished the use pipeline')


#!/usr/bin/env python
# Imports
import unittest
import os.path
import sys
import subprocess

###############################################################################

path_to_script = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','enrichm')
path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path

###############################################################################

class Tests(unittest.TestCase):

    def test_hello_world(self):
        cmd = '%s -h > /dev/null' % path_to_script
        subprocess.call(cmd, shell=True)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import logging
import sys
import os
import shutil
import time

from enrichm.data import Data
from enrichm.network_analyzer import NetworkAnalyser
from enrichm.enrichment import Enrichment
from enrichm.annotate import Annotate
from enrichm.classifier import Classify
from enrichm.generate import GenerateModel
from enrichm.predict import Predict
from enrichm.uses import Uses

####################################################################################################

debug = {1:logging.CRITICAL, 2:logging.ERROR, 3:logging.WARNING, 4:logging.INFO, 5:logging.DEBUG}

####################################################################################################

class Run:

    def __init__(self):

        self.DATA = 'data'
        self.ANNOTATE = 'annotate'
        self.CLASSIFY = 'classify'
        self.ENRICHMENT = 'enrichment'
        self.PREDICT = 'predict'
        self.GENERATE = 'generate'
        self.USES = 'uses'

    def _logging_setup(self, args):
        if args.verbosity not in range(1, 6):
            raise Exception("Logging verbosity must be a positive integer between 1 and 5.")

        logger = logging.getLogger('')
        logger.setLevel(debug[args.verbosity])
        log_format = logging.Formatter(fmt="[%(asctime)s] %(levelname)s: %(message)s",
                                       datefmt="%Y-%m-%d %H:%M:%S %p")

        stream_logger = logging.StreamHandler(sys.stdout)
        stream_logger.setFormatter(log_format)
        stream_logger.setLevel(debug[args.verbosity])
        logger.addHandler(stream_logger)

        if args.subparser_name!=self.DATA:
            file_logger = logging.FileHandler(os.path.join(args.output, args.log), 'a')
            file_logger.setFormatter(log_format)
            logger.addHandler(file_logger)

    def _check_general(self, args):
        '''
        Check general input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        dependencies = {'hmmsearch':"http://hmmer.org/download.html",
                        'diamond':"https://github.com/bbuchfink/diamond",
                        'R':"https://www.r-project.org",
                        'parallel':"https://www.gnu.org/software/parallel",
                        'prodigal':"https://github.com/hyattpd/Prodigal/wiki/installation",
                        'mmseqs':"https://github.com/soedinglab/MMseqs2"}

        missing_dependencies = list()

        for dependency in dependencies.keys():

            if shutil.which(dependency) == None:
                missing_dependencies.append(dependency)

        if len(missing_dependencies)>0:
            dependency_string = '\n'.join(['\t%s\t%s' % (d, dependencies[d]) for d in missing_dependencies])
            raise Exception('The following dependencies need to be installed to run enrichm:\n%s' % (dependency_string))

        if args.subparser_name != self.DATA:
            # Set up working directory
            if not args.output:
                args.output = '%s-enrichm_%s_output' % (time.strftime("%Y-%m-%d_%H-%M"), args.subparser_name)

            if(os.path.isdir(args.output) or os.path.isfile(args.output)):

                if args.force:

                    if os.path.isdir(args.output):
                        shutil.rmtree(args.output)

                    else:
                        os.remove(args.output)

                else:
                    raise Exception("File '%s' exists." % args.output)

            os.mkdir(args.output)

    def _check_data(self, args):
        if not(args.create or args.uninstall):
            raise Exception("Only one of the following can be specified: --create, --uninstall")

    def _check_annotate(self, args):
        '''
        Check annotate input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        # ensure either a list of genomes or a directory of genomes have been specified
        if not(args.genome_files or args.genome_directory or args.protein_directory or args.protein_files):
            raise Exception("Input error: Either a list of genomes or a directory of genomes need to be specified.")

        if len([x for x in [args.genome_files, args.genome_directory, args.protein_directory, args.protein_files] if x]) != 1:
            raise Exception("Input error: Only one type of input can be specified (--genome_files, --genome_directory, --protein_directory, or --protein_files).")

        if not args.suffix:

            if(args.genome_directory or args.genome_files):
                args.suffix = '.fna'

            elif(args.protein_directory or args.protein_files):
                args.suffix = '.faa'

        if(args.id>1 or args.id<0):
            raise Exception("Identity (--id) must be between 0 and 1.")

        if(args.aln_query>1 or args.aln_query<0):
            raise Exception("Alignment to query cutoff (--aln_query) must be between 0 and 1")

        if(args.aln_reference>1 or args.aln_reference<0):
            raise Exception("Alignment to reference cutoff (--aln_reference) must be between 0 and 1")

        if any([args.cut_ga_pfam, args.cut_nc_pfam, args.cut_tc_pfam]):
            if len([x for x in [args.cut_ga_pfam, args.cut_nc_pfam, args.cut_tc_pfam] if x])>1:
                raise Exception("Only one of the following can be selected: --cut_ga_pfam, --cut_nc_pfam, --cut_tc_pfam")
            if args.evalue:
                logging.warning('selecting one of the following overrides evalue thresholds: --cut_ga_pfam, --cut_nc_pfam, --cut_tc_pfam, --cut_ga_tigrfam, --cut_nc_tigrfam, --cut_tc_tigrfam')

        if any([args.cut_ga_tigrfam, args.cut_nc_tigrfam, args.cut_tc_tigrfam]):
            if len([x for x in [args.cut_ga_tigrfam, args.cut_nc_tigrfam, args.cut_tc_tigrfam] if x])>1:
                raise Exception("Only one of the following can be selected: --cut_ga_tigrfam, --cut_nc_tigrfam, --cut_tc_tigrfam")
            if args.evalue:
                logging.warning('selecting one of the following overrides evalue thresholds: --cut_ga_pfam, --cut_nc_pfam, --cut_tc_pfam, --cut_ga_tigrfam, --cut_nc_tigrfam, --cut_tc_tigrfam')

    def _check_enrichment(self, args):
        '''
        Check enrichment input and output options are valid.

        Parameters

        args    - object. Argparse object

        Output
        ------
        '''
        ### ~ TODO: Check Multi test correction inputs...
        types = [args.ko, args.pfam, args.tigrfam, args.cluster, args.ortholog, args.cazy, args.ec, args.ko_hmm]

        if not args.abundance and args.abundance_metadata:
           raise Exception("Values for both --abundance and --abundance_metadata are required")
        if not args.transcriptome and args.transcriptome_metadata:
           raise Exception("Values for both --abundance and --abundance_metadata are required")
        if args.annotation_matrix and args.annotate_output:
            raise Exception("Use either --annotate_output or --annotation_matrix")

        if(not args.annotation_matrix and not args.annotate_output and not args.gff_files):
            raise Exception("Either --annotate_output, --annotation_matrix or --gff_files must be specified!")

        if args.annotation_matrix or args.annotate_output:
            if not args.abundance:
                if not args.metadata:
                    raise Exception("Genome groups need to be specified using the --metadata flag")

        if args.annotate_output:

            if not any(types):
                raise Exception(
                    "Input Error: One of the following flags must be specified: --ko --pfam --tigrfam --orthologs --orthogroup --clusters --ko_hmm --ec --cazy")

            if len([x for x in types if x]) > 1:
                raise Exception(
                    "Only one of the following flags may be specified: --ko --pfam --tigrfam --orthologs --orthogroup --clusters --ko_hmm --ec --cazy")

    def _check_classify(self, args):
        '''
        Check classify input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        if(args.cutoff>1 and args.cutoff<0):
            raise Exception("--cutoff needs to be between 0 and 1")

        if(args.gff_files and args.genome_and_annotation_matrix):
            raise Exception(f"Both --gff_files {args.gff_files} and --genome_and_annotation_matrix {args.genome_and_annotation_matrix} were specified. Please provide only one option.")

        if(args.module_rules_json and not args.gff_files):
            raise Exception(f"--gff_files must be provided to use --module_rules_json")

        if args.module_rules_json:
            if not os.path.isfile(args.module_rules_json):
                raise Exception(f"File does not exist: {args.module_rules_json}")

        if args.custom_modules:
            if not os.path.isfile(args.custom_modules):
                raise Exception(f"File does not exist: {args.custom_modules}")

    def _check_network(self, args):
        '''
        Check network (explore, pathway) input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        if not hasattr(args, 'enrichment_output'):
            args.enrichment_output = None

        if any([args.abundance, args.abundance_metadata]):
            if not (args.abundance and args.abundance_metadata):
                raise Exception("Both abundance and abundance metadata need to be specified")

        if any([args.tpm_values, args.tpm_metadata]):
            if not (args.tpm_values and args.tpm_metadata):
                raise Exception("Both --tpm_values and --tpm_metadata need to be specified")

        if args.subparser_name == NetworkAnalyser.PATHWAY:
            args.depth = None
            args.queries = None

        if args.subparser_name == NetworkAnalyser.EXPLORE:
            args.filter = None
            args.limit = None

            if not(args.queries):

                if args.depth:
                    logging.warning("--depth argument ignored without --queries flag")

    def _check_predict(self, args):
        '''
        Check general input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        pass

    def _check_uses(self, args):
        '''
        Check general input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        pass

    def _check_generate(self, args):
        '''
        Check general input and output options are valid.

        Parameters
        ----------
        args    - object. Argparse object
        '''
        pass

    def run_data(self, args):
        d = Data()
        d.do(args.uninstall, args.create, args.dry)

    def run_annotate(self, args):
        self._check_annotate(args)

        annotate = Annotate(# Define inputs and outputs
                            args.output,
                            # Define type of annotation to be carried out
                            args.ko, args.ko_hmm, args.pfam, args.tigrfam,
                            args.clusters, args.orthologs, args.cazy,
                            args.ec, args.orthogroup,
                            # Cutoffs
                            args.evalue, args.bit, args.id, args.aln_query,
                            args.aln_reference, args.c, args.cut_ga_pfam,

                            args.cut_nc_pfam, args.cut_tc_pfam,
                            args.cut_ga_tigrfam, args.cut_nc_tigrfam,
                            args.cut_tc_tigrfam, args.cut_ko,
                            args.inflation, args.chunk_number, args.chunk_max,
                            args.count_domains,
                            # Parameters
                            args.threads, args.parallel, args.suffix, args.light)

        annotate.annotate_pipeline(args.genome_directory,
                                    args.protein_directory,
                                    args.genome_files,
                                    args.protein_files)

    def run_classify(self, args):
        self._check_classify(args)
        classify = Classify()
        classify.classify_pipeline(args.custom_modules, args.cutoff, args.aggregate,
                                    args.genome_and_annotation_matrix, args.module_rules_json, 
                                    args.gff_files, args.output)

    def run_enrichment(self, args):
        self._check_enrichment(args)
        enrichment = Enrichment()
        enrichment.enrichment_pipeline(# Input options
                                        args.annotate_output, args.annotation_matrix, args.gff_files,
                                        args.metadata, args.abundance, args.abundance_metadata,
                                        args.transcriptome, args.transcriptome_metadata,
                                        # Runtime options
                                        args.pval_cutoff, args.proportions_cutoff, 
                                        args.threshold, args.multi_test_correction, 
                                        args.batchfile, args.processes, 
                                        args.allow_negative_values, args.ko, args.pfam, 
                                        args.tigrfam, args.cluster, args.ortholog, args.cazy,
                                        args.ec, args.ko_hmm, args.range, args.subblock_size,
                                        args.operon_mismatch_cutoff, args.operon_match_score_cutoff,
                                        # Outputs
                                        args.output)


    def run_network(self, args):
        self._check_network(args)
        network_analyser=NetworkAnalyser()
        network_analyser.network_pipeline(args.subparser_name, args.matrix, 
                                            args.genome_metadata, args.tpm_values,
                                            args.tpm_metadata, args.abundance, 
                                            args.abundance_metadata, args.metabolome,
                                            args.enrichment_output, args.depth, args.filter,
                                            args.limit, args.queries, args.output)


    def run_predict(self, args):
        self._check_predict(args)
        predict = Predict()
        predict.predict_pipeline(args.forester_model_directory,
                args.input_matrix,
                args.output)

    def run_generate(self, args):
        self._check_generate(args)
        generate_model = GenerateModel()
        generate_model.generate_pipeline(args.input_matrix,
                args.groups,
                args.model_type,
                args.testing_portion,
                args.grid_search,
                args.threads,
                args.output)


    def run_uses(self, args):
        self._check_uses(args)
        uses = Uses()
        uses.uses_pipeline(args.compounds_list,
                args.annotation_matrix,
                args.metadata,
                args.output,
                args.count)

    def get_pipeline(self, subparser_name):
        if subparser_name == self.DATA:
            pipeline = self.run_data
        elif subparser_name == self.ANNOTATE:
            pipeline = self.run_annotate
        elif subparser_name == self.CLASSIFY:
            pipeline = self.run_classify
        elif subparser_name == self.ENRICHMENT:
            pipeline = self.run_enrichment
        elif subparser_name == NetworkAnalyser.PATHWAY:
            pipeline = self.run_network
        elif subparser_name == NetworkAnalyser.EXPLORE:
            pipeline = self.run_network
        elif subparser_name == self.PREDICT:
            pipeline = self.run_predict
        elif subparser_name == self.GENERATE:
            pipeline = self.run_generate
        elif subparser_name == self.USES:
            pipeline = self.run_uses
        else:
            ValueError(subparser_name)


        return pipeline

    def run_enrichm(self, args, command):
        '''
        Parameters
        ----------

        Output
        ------
        '''
        self._check_general(args)
        self._logging_setup(args)

        logging.info("Command: %s" % ' '.join(command))
        logging.info("Running the %s pipeline" % args.subparser_name)
        pipeline = self.get_pipeline(args.subparser_name)
        pipeline(args)
        logging.info('Finished running EnrichM')



#!/usr/bin/env python
# Imports
import unittest
import os.path
import sys
import subprocess
import tempfile

###############################################################################

path_to_script = os.path.join(os.path.dirname(os.path.realpath(__file__)),'..','bin','enrichm')
path_to_data = os.path.join(os.path.dirname(os.path.realpath(__file__)),'data')
path_to_annotate = os.path.join(path_to_data, 'enrichm_annotate')
sys.path = [os.path.join(os.path.dirname(os.path.realpath(__file__)),'..')]+sys.path

from enrichm.classifier import Classify
from enrichm.module_description_parser import ModuleDescription

###############################################################################

class Tests(unittest.TestCase):

    def test_classify_from_matrix(self):
        tmp = tempfile.mkdtemp()
        
        ko_matrix = os.path.join(path_to_annotate, 'ko_frequency_table.tsv')
        cmd = f'{path_to_script} classify \
                    --genome_and_annotation_matrix {ko_matrix} \
                    --output {tmp} \
                    --force \
                    --verbosity 1'
        result = subprocess.check_call(cmd, shell=True)

    def test_classify_with_custom_modules(self):
        tmp = tempfile.mkdtemp()
        ko_matrix = os.path.join(path_to_annotate, 'ko_frequency_table.tsv')
        cmd = f'{path_to_script} classify \
                    --genome_and_annotation_matrix {ko_matrix} \
                    --output {tmp} \
                    --force \
                    --verbosity 1'
        subprocess.check_call(cmd, shell=True)

    def test_curate(self):
        # M00516 
        module = ModuleDescription("K11231,K19690,K19691,K19692 K11232 K11233,K15859")
        total_steps = module.num_steps()
        fails1 = ['K11231', 'K11232']
        passes1 = ['K11231', 'K11232', 'K11233']
        self.assertEqual(False,
                         module.num_covered_steps(fails1)[0] == total_steps)
        self.assertEqual(True,
                         module.num_covered_steps(passes1)[0] == total_steps)
        # M00156
        module = ModuleDescription("((K00404+K00405,K15862)+K00407+K00406)")
        total_steps = module.num_steps()
        fails1 = ['K15862', 'K00406']
        fails2 = ['K00404', 'K00407', 'K00406']
        passes1 = ['K00404', 'K00405', 'K00407', 'K00406']
        passes2 = ['K15862', 'K00407', 'K00406']
        self.assertEqual(False,
                         module.num_covered_steps(fails1)[0] == total_steps)
        self.assertEqual(True,
                         module.num_covered_steps(passes1)[0] == total_steps)
        self.assertEqual(False,
                         module.num_covered_steps(fails2)[0] == total_steps)
        self.assertEqual(True,
                         module.num_covered_steps(passes2)[0] == total_steps)

        # M00006
        module = ModuleDescription("(K13937,((K00036,K19243) (K01057,K07404))) K00033")
        total_steps = module.num_steps()
        fails1 = ['K19243', 'K00033', 'K19243']
        fails2 = ['K19243', 'K00033']
        passes1 = ['K00036', 'K01057', 'K00033']
        passes2 = ['K13937', 'K00033']
        passes3 = ['K19243', 'K07404', 'K00033']
        passes4 = ['K00036', 'K07404', 'K00033']

        self.assertEqual(True,
                         module.num_covered_steps(passes1)[0] == total_steps)
        self.assertEqual(True,
                         module.num_covered_steps(passes2)[0] == total_steps)
        self.assertEqual(True,
                         module.num_covered_steps(passes3)[0] == total_steps)
        self.assertEqual(True,
                         module.num_covered_steps(passes4)[0] == total_steps)
        self.assertEqual(False,
                         module.num_covered_steps(fails1)[0] == total_steps)
        self.assertEqual(False,
                         module.num_covered_steps(fails2)[0] == total_steps)


    def test_update(self):

        with open(tempfile.mktemp(), 'w') as out_io:
            out_io.write(f"test\tcontents\n")
            out_io.flush()
            classify = Classify()
            classify.update_with_custom_modules(out_io.name)
            self.assertEqual(True, 'test' in classify.modules)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=line-too-long
'''
Check for blocks of enriched genes that co-occur within the genome.
'''
import logging
from itertools import product, chain, combinations
from fuzzywuzzy import fuzz, process
from collections import Counter
from enrichm.toolbox import reverse_dictionary_of_lists, cluster

class SyntenySearcher:

    def __init__(self):
        self.synteny_results_output_file = "synteny_results.tsv"
        self.synteny_results_header = ["Genome_group", "Core_gene_block", "Num_genes", "Group_genomes", "Hit_genomes", "Percent_genomes",  "Genomes"]


    def cluster_operons_by_common_elements(self, synteny_to_genome, operon_mismatch_cutoff, operon_match_score_cutoff):
        cluster_list = list()
        
        for reference_operon in synteny_to_genome:
            operon_cluster = set([reference_operon])
            reference_operon_set = set(reference_operon.split('~'))
            reference_operon_length = len(reference_operon_set)

            for query_operon in synteny_to_genome:
                  
                if query_operon not in reference_operon:
                    query_operon_set = set(query_operon.split('~'))
                    query_operon_length = len(query_operon_set)
                    overlap_length = len(reference_operon_set.intersection(query_operon_set))
                    operon_match_score = overlap_length / reference_operon_length
                    mismatch_count = query_operon_length - overlap_length
                    
                    
                    if operon_match_score >= operon_match_score_cutoff:

                        if mismatch_count <= operon_mismatch_cutoff:
                            operon_cluster.add(query_operon)
            
            # Add if not duplicate or subset of another cluster list
            if len(cluster_list) == 0:
                cluster_list.append(operon_cluster)
            else:
                new = True
                
                for previous_operon_cluster in cluster_list:
                    previous_cluster_length = len(previous_operon_cluster)
                    current_cluster_length = len(operon_cluster)
                    overlap_length = len(previous_operon_cluster.intersection(operon_cluster))

                    if(overlap_length==previous_cluster_length or overlap_length==current_cluster_length):
                        new = False
                
                if new:
                    cluster_list.append(operon_cluster)
        
        #for synteny_block_a, synteny_block_b in combinations(synteny_to_genome.keys(), 2):
        #    synteny_block_a_set = set(synteny_block_a.split('~'))
        #    synteny_block_b_set = set(synteny_block_b.split('~'))
        #    synteny_overlap = set(synteny_block_a_set).intersection(set(synteny_block_b_set))
        #    
        #    if len(synteny_overlap)>=2: 
        #        # Quick and dirtly implementation. If two synteny blocks share >=2 common elements,
        #        # cluster them together ALTHOUGH this isnt a very effective method to protect against
        #        # cluster creep, which is dangerous.
        #        found = False

        #        for cl in cluster_list:
        #            
        #            if(synteny_block_a in cl or synteny_block_b in cl):
        #                core = set.intersection(*[set(x.split('~')) for x in cl])

        #                if(len(synteny_block_a_set.intersection(core))>=2 and len(synteny_block_b_set.intersection(core))>=2):
        #                    cl.update([synteny_block_a, synteny_block_b])
        #                    grouped_clusters.update([synteny_block_a, synteny_block_b])
        #                    found = True

        #        if not found:
        #            cluster_list.append(set([synteny_block_a, synteny_block_b]))
        #            grouped_clusters.update([synteny_block_a, synteny_block_b])
        #
        ## Add in ungrouped operons
        #cluster_list += [set([cluster]) for cluster in synteny_to_genome if cluster not in grouped_clusters]
        return cluster_list


    def search_for_blocks(self, prevalence_results, gene_positions, metadata, synteny_range, min_subblock_size,
                          operon_mismatch_cutoff, operon_match_score_cutoff):
        """
        Identify groups of significantly enriched genes that are also co-located between groups of genomes
    
        :type prevalence_results: list
        :param prevalence_results:
    
        :type gene_positions: dict
        :param gene_positions:
    
        :type metadata: dict
        :param metadata:
    
        :type synteny_range: int
        :param synteny_range:
    
        :type min_subblock_size: int
        :param min_subblock_size:
    
        :raises:
    
        :rtype:
        """
        group_to_enriched_gene = dict()
        metadata_reverse_mapping = reverse_dictionary_of_lists(metadata)
        output_lines = [self.synteny_results_header]

        # TODO This for loop can be functionized
        for result in prevalence_results[1:]:
            gene = result[0]
            enriched_in = result[3]
            corrected_pvalue = float(result[-2])

            # Just ignore genes that arent enriched on one group
            # for now
            if enriched_in == 'NA':
                continue

            # Ignore genes that aren't significantly different after multiple
            # test correction
            if corrected_pvalue > 0.05:
                continue
            
            if enriched_in not in group_to_enriched_gene:
                group_to_enriched_gene[enriched_in] = list()

            group_to_enriched_gene[enriched_in].append(gene)
        
        for group, enriched_gene_list in group_to_enriched_gene.items():
            synteny_to_genomes = dict()

            for genome in metadata_reverse_mapping[group]:
                
                for contig, genome_positions in gene_positions[genome].items(): 
                    # This is iterating through a genome on a contig-by-contig basis
                    # with nothing in place to allow operons to be identified across cotnigs
                    # This is stringent, but maybe necessary?
                    
                    endings_dict = dict()

                    for gene in enriched_gene_list:

                        if gene in genome_positions:
                            for pos in genome_positions[gene]:
                                endings_dict[pos[1]] = gene 
                        else:
                            logging.debug(f"Enriched gene {gene} not found on contig {contig} in {genome}. This situation is ignored.")

                    if len(endings_dict):

                        for subblock in cluster(list(endings_dict.keys()), synteny_range):

                            if len(subblock)>=min_subblock_size:
                                block = [endings_dict[bit] for bit in subblock]
                                synteny_definition = '~'.join(block)

                                if synteny_definition not in synteny_to_genomes:
                                    synteny_to_genomes[synteny_definition] = list()

                                synteny_to_genomes[synteny_definition].append(genome)
                    else:
                        logging.debug(f"No syntenous blocks found for {genome} on contig {contig}")
            
            clustered_operons = self.cluster_operons_by_common_elements(synteny_to_genomes, operon_mismatch_cutoff, operon_match_score_cutoff)
            
            for operon_group in clustered_operons:
                core_genes = '~'.join(set.intersection(*[set(x.split('~')) for x in operon_group]))
                found_genomes = set(chain(*[synteny_to_genomes[x] for x in operon_group]))
                perc_genomes = round(len(found_genomes) / len(metadata_reverse_mapping[group])*100, 2)
                output_lines.append([group, core_genes, len(core_genes.split('~')), len(metadata_reverse_mapping[group]), len(found_genomes), perc_genomes, "~".join(found_genomes)])

        return output_lines, self.synteny_results_output_file